from brainannex.neo_schema.neo_schema import NeoSchema


class Recordsets:
    """
    Plugin-provided custom interface for "recordsets"
    """

    SCHEMA_CLASS_NAME = "Recordset"



    @classmethod
    def initialize_schema(cls) -> None:
        """
        Initialize the Schema needed by this plugin

        :return:    None
        """
        assert NeoSchema.is_valid_class_name(cls.SCHEMA_CLASS_NAME), \
            f"initialize_schema(): attempting to create a Schema Class with an invalid name: '{cls.SCHEMA_CLASS_NAME}'"

        if not NeoSchema.class_name_exists("Content Item"):
            NeoSchema.create_class_with_properties(name="Content Item", strict=True,
                                               properties=["uri"])

        db_id, _ = NeoSchema.create_class_with_properties(name=cls.SCHEMA_CLASS_NAME, strict=True, code="rs",
                                                          properties=["class", "order_by", "clause", "n_group", "caption"],
                                                          class_to_link_to="Content Item", link_name="INSTANCE_OF", link_dir="OUT")

        # Set data types for some Properties
        NeoSchema.set_property_attribute(class_name=cls.SCHEMA_CLASS_NAME, prop_name="n_group",
                                         attribute_name="dtype", attribute_value="int")


        # Set up the auto-increment namespace
        NeoSchema.create_namespace(name="recordset", prefix="rs-", suffix="")
        match_to = NeoSchema.db.match(labels="Schema Autoincrement", key_name="namespace", key_value="recordset")
        NeoSchema.db.add_links(match_from=db_id, match_to=match_to, rel_name="HAS_URI_GENERATOR")



from typing import Union, List
from brainannex.neo_schema.neo_schema import NeoSchema
from neoaccess import NeoAccess


class Collections:
    """
    An ordered sequence of Data Nodes.

    We define a "Collection" as an entity to which a variety of Data Nodes
    are attached, with positional attributes in their links.

    In terms of implementation, it's a hub, with a central data node (the "Collection"),
    to which any number of Data Nodes (or "Collection Items") are linked;
    the relationship name can be anything (user-passed) but it gets assigned a property
    named "pos", which is managed by this class, to maintain and alter the sequential order.

    Example of use case: "pages" of "Content Items" attached to a "Category"
    """
    # TODO: maybe rename the class to OrderedCollections (or OrderedSequences)

    # Class variables

    db = None                       # MUST be set before using this class!

    DELTA_POS = 20                  # Arbitrary default shift in "pos" value between Collection Items that are adjacent
                                    # (i.e. next in sequence);
                                    # empirically shown best to be even, and not too small nor too large

    membership_rel_name = None      # NOT IN USE.   TODO: maybe use instantiation for this class, and set at that time



    @classmethod
    def set_database(cls, db :NeoAccess) -> None:
        """
        IMPORTANT: this method MUST be called before using this class!

        :param db:  Database-interface object, created with the NeoAccess library
        :return:    None
        """

        assert type(db) == NeoAccess, \
            "Collections.set_database(): argument passed isn't a valid `NeoAccess` object"

        cls.db = db



    @classmethod
    def initialize_collections(cls) -> (int, str):
        """
        Create a new Schema Class node that represents a "Collection"

        :return:    An (int, str) pair of integers with the internal database ID
                        and the unique uri assigned to the node just created
        """
        return NeoSchema.create_class(name="Collections", strict = False)



    @classmethod
    def is_collection(cls, collection_uri :str) -> bool:
        """
        Return True if the Data Node with the given uri is a Collection,
        that is, if its schema is a Class that is an INSTANCE_OF the "Collections" Class

        :param collection_uri:  A string with the URI of a Data Node
        :return:                True if the given data node is a Collection, or False otherwise
        """

        #TODO: maybe allow the scenario where there's a longer chain of "INSTANCE_OF" relationships??
        q = '''
            MATCH p=({uri: $collection_uri}) -[:SCHEMA]-> (:CLASS) 
                    -[:INSTANCE_OF]-> 
                    (:CLASS {name: "Collections"})
            RETURN count(p) AS number_paths
            '''
        data_binding = {"collection_uri": collection_uri}
        number_paths = cls.db.query(q, data_binding, single_cell="number_paths")

        return True if number_paths > 0 else False



    @classmethod
    def collection_size(cls, collection_id :str, membership_rel_name: str, skip_check=False) -> int:
        """
        Return the number of elements in the given Collection (i.e. Data Items linked to it thru the specified relationship)

        :param collection_id:       The uri of a data node whose schema is an instance of the Class "Collections"
        :param membership_rel_name: The name of the relationship from other Data Items to the given Collection node
        :param skip_check:          If True, no check is done to verify that the data node whose uri matches collection_id
                                    is indeed a Collection.
                                    Without a check, this function will return a zero if given a bad collection_id;
                                    with a check, it'll raise an Exception
        :return:                    The number of elements in the given Collection (possibly zero)
        """
        if not skip_check:
            assert cls.is_collection(collection_id), \
                    f"The data node with uri `{collection_id}` doesn't exist or is not a Collection"

        q = f'''
            MATCH ({{uri: $collection_id}}) <- [:{membership_rel_name}] - (i) 
            RETURN count(i) AS node_count
            '''
        data_binding = {"collection_id": collection_id}

        number_items_attached = cls.db.query(q, data_binding, single_cell="node_count")
        return number_items_attached



    @classmethod
    def delete_collection_NOT_IMPLEMENTED(cls, collection_uri :str) -> None:
        """
        #TODO: implement

        Delete the specified Collection, provided that there are no Data Items linked to it.
        In case of error or failure, an Exception is raised.

        :param collection_uri:  The "uri" string value identifying the desired Collection
        :return:                None
        """
        pass



    @classmethod
    def add_to_collection_at_beginning(cls, collection_uri :str, membership_rel_name: str, item_class_name: str, item_properties: dict,
                                       new_uri=None) -> str:
        """
        Create a new data node, of the class specified in item_class_name, and with the given properties -
        and add it at the beginning of the specified Collection, linked by the specified relationship

        EXAMPLE:  new_uri = add_to_collection_at_beginning(collection_id=708, membership_rel_name="BA_in_category",
                                                        item_class_name="Headers", item_properties={"text": "New Caption, at the end"})
        <SEE add_to_collection_at_end>

        :return:                    The auto-increment "uri" assigned to the newly-created data node
        """
        #TODO:  the Data Node creation perhaps should be done ahead of time, and doesn't need to be the responsibility of this class!
        #TODO:  solve the concurrency issue - of multiple requests arriving almost simultaneously, and being handled by a non-atomic update,
        #       which can lead to incorrect values of the "pos" relationship attributes.
        #       -> Follow the new way it is handled in link_to_collection_at_end()

        assert NeoSchema.is_valid_uri(collection_uri), "The argument `collection_uri` isn't a valid URI string"
        assert type(membership_rel_name) == str, "The argument `membership_rel_name` MUST be a string"
        assert type(item_class_name) == str, "The argument `item_class_name` MUST be a string"
        assert type(item_properties) == dict, "The argument `item_properties` MUST be a dictionary"

        # TODO: this query and the one in add_data_point(), below, ought to be combined, to avoid concurrency problems
        q = f'''
            MATCH (n:BA) - [r :{membership_rel_name}] -> (c:BA {{uri: $collection_id}}) 
            RETURN min(r.pos) AS min_pos
            '''
        data_binding = {"collection_id": collection_uri}

        min_pos = cls.db.query(q, data_binding, single_cell="min_pos")  # This will be None if the Collection has no elements

        if min_pos is None:
            pos = 0                         # Arbitrary initial value for cases when the Collection has no elements
        else:
            pos = min_pos - cls.DELTA_POS   # Go some distance before the beginning

        data_binding = item_properties

        return NeoSchema.add_data_point_OLD(class_name=item_class_name,
                                            data_dict=data_binding,
                                            labels=["BA", item_class_name],
                                            connected_to_id=collection_uri, connected_to_labels="BA",
                                            rel_name=membership_rel_name,
                                            rel_prop_key="pos", rel_prop_value=pos,
                                            new_uri=new_uri
                                            )



    @classmethod
    def link_to_collection_at_end(cls, item_uri :str, collection_uri :str, membership_link_name :str) -> None:
        """
        Given an existing Data Node (regarded as a "Collection Item"),
        link it to the end of the specified Collection data node,
        using the requested relationship name.

        If a link already exists, an Exception is raised.

        Note: a database lock is used to deal with multiple concurrent calls.

        :param item_uri:            The URI of an existing Data Node representing a "Collection Item"
        :param collection_uri:      The URI of an existing Collection Data Node
        :param membership_link_name:The name to give to the new relationship to be created
                                        in the direction from the "Collection Item" to the Collection node
        :return:                    None
        """
        # TODO: use this function as a ***ROLE MODEL*** for linking at the start, or after an existing Item
        # TODO: (optionally?) enforce that the respective Classes of the Data Nodes are a relationship named membership_rel_name;
        #       use NeoSchema.class_relationship_exists()

        # TODO: validate more thoroughly the URI's
        assert item_uri, \
            "link_to_collection_at_end: Missing `item_uri` argument"

        assert collection_uri, \
            "link_to_collection_at_end: Missing `collection_uri` argument"

        assert membership_link_name, \
            "link_to_collection_at_end: Missing `item_uri` membership_rel_name"

        # Database update, with a DATA LOCK to protect against multiple concurrent calls,
        # that locates the next-available "pos" number, and creates a relationship using it.
        # Info on using a lock this way: https://neo4j.com/docs/java-reference/4.4/transaction-management/
        # The "OPTIONAL MATCH" is used to compute a new positional value
        q = f'''
            MATCH (ci), (collection) 
            WHERE ci.uri = $item_uri 
              AND collection.uri = $collection_uri
              AND NOT ( (ci) -[:`{membership_link_name}`]-> (collection) )
            SET collection._LOCK_ = true
            WITH ci, collection
            
            OPTIONAL MATCH (existing_ci) -[r :`{membership_link_name}`]-> (collection)
            WITH r.pos AS pos, collection, ci
            WITH 
                CASE WHEN pos IS NULL THEN
                    0
                ELSE
                    max(pos) + {cls.DELTA_POS}
                END AS new_pos, collection, ci
            
            MERGE (ci) -[:`{membership_link_name}` {{pos: new_pos}}]-> (collection)
            
            REMOVE collection._LOCK_
            '''

        data_binding={"collection_uri": collection_uri, "item_uri": item_uri}

        #cls.db.debug_query_print(q, data_binding=data_binding)

        status = cls.db.update_query(q, data_binding=data_binding)
        #print("link_to_collection_at_end(): status is ", status)
        # status should be contain {'relationships_created': 1, 'properties_set': 3}

        # NOTE: the 3 derives from the "pos" attribute set on the newly-created link,
        #       plus the 2 temp properties set and cleared for the lock
        assert status.get('properties_set') == 3, \
            f"link_to_collection_at_end(): failed to set the positional value to the new link " \
            f"to the Collection with URI '{collection_uri}'"

        assert status.get('relationships_created') == 1, \
            f"link_to_collection_at_end(): failed to create a new link " \
            f"to a Collection with URI '{collection_uri}'"



    @classmethod
    def relocate_to_other_collection_at_end(cls, item_uri :str, from_collection_uri :str, to_collection_uri :str, membership_rel_name :str):
        """
        Given an existing data node (representing a "Collection Item" of the specified "from" Collection),
        switch it to become a "Collection Item" of the "to" Collection, positioned at the end of it.

        The collection-membership relationship is severed from the "Collection Item" to the "from" Collection,
        and a new one is created from the "Collection Item" to the "to" Collection.

        In case no operation is performed, an Exception is raised.

        :param item_uri:            The URI of a Data Node representing a "Collection Item" of the "from" Collection below
        :param from_collection_uri: The URI of a Collection Data Node to which the above "Collection Item" is connected
        :param to_collection_uri:   The URI of a Collection Data Node to which the above "Collection Item" needs to be switched to
        :param membership_rel_name: The name to give to the relationship
                                        in the direction from the "Collection Item" to the Collection node
        :return:                    None
        """
        # TODO: perhaps to ditch, now that we have bulk_relocate_to_other_collection_at_end()

        # Use an ATOMIC operation.  If any of the matches fail, no operation is performed
        # The "OPTIONAL MATCH" is used to compute a new positional value
        q = f'''
            MATCH (collection_from) , (collection_to) ,
                  (moving_ci) -[old_r :`{membership_rel_name}`]-> (collection_from)
            WHERE collection_from.uri = $from_collection_uri
              AND collection_to.uri = $to_collection_uri
              AND moving_ci.uri = $item_uri
            WITH collection_from, collection_to, moving_ci, old_r
                        
            OPTIONAL MATCH (existing_ci) -[r :`{membership_rel_name}`]-> (collection_to)
            WITH r.pos AS pos, collection_from, collection_to, moving_ci, old_r
            WITH 
                CASE WHEN pos IS NULL THEN
                    0
                ELSE
                    max(pos) + {cls.DELTA_POS}
                END AS new_pos, collection_from, collection_to, moving_ci, old_r
            
            DELETE old_r
            MERGE (moving_ci)-[:`{membership_rel_name}` {{pos: new_pos}}]->(collection_to)
            '''
        #cls.db.debug_query_print(q, data_binding={"from_collection_uri": from_collection_uri,
                                                  #"to_collection_uri": to_collection_uri,
                                                  #"item_uri": item_uri})

        status = cls.db.update_query(q,
                                     data_binding={"from_collection_uri": from_collection_uri,
                                                   "to_collection_uri": to_collection_uri,
                                                   "item_uri": item_uri})
        #print("switch_to_other_collection_at_end(): status is ", status)

        # status should contain {'relationships_deleted': 1, 'relationships_created': 1, 'properties_set': 1}

        assert status.get('relationships_deleted') == 1, \
            f"relocate_to_other_collection_at_end(): failed to locate or delete the old '{membership_rel_name}' link " \
            f"that goes to a Collection with URI '{from_collection_uri}'"

        assert status.get('relationships_created') == 1, \
            f"relocate_to_other_collection_at_end(): failed to create a new link " \
            f"to a Collection with URI '{to_collection_uri}'"

        assert status.get('properties_set') == 1, \
            f"relocate_to_other_collection_at_end(): failed to set the positional value to the new link " \
            f"to a Collection with URI '{to_collection_uri}'"



    @classmethod
    def bulk_relocate_to_other_collection_at_end(cls, items :Union[List[str], str],
                                                 from_collection :str, to_collection :str, membership_rel_name :str) -> int:
        """
        Given an existing list of data nodes (representing "Collection Items" of the specified "from" Collection),
        switch each of them to become a "Collection Item" of the "to" Collection, positioned at the end of it.

        The collection-membership relationship is severed from each of the "Collection Items" to the "from" Collection,
        and a new one is created from that "Collection Item" to the "to" Collection.

        Return the number of Collection Items successfully relocated.

        :param items:               URI, or list of URI's, of Data Node(s)
                                        representing a "Collection Items" of the "from" Collection below
        :param from_collection:     The URI of a Collection Data Node to which the above Collection Item(s) are connected
        :param to_collection:       The URI of a Collection Data Node to which the above Collection Item(s) needs to be switched to
        :param membership_rel_name: The name to give to the relationship
                                        in the direction from the "Collection Item" to the Collection node
        :return:                    The number of Collection Items successfully relocated
        """
        # If a scalar was passed, turn into a list
        if type(items) == str:
            items = [items]
        else:
            assert type(items) == list, \
                "bulk_relocate_to_other_collection_at_end(): the argument `items` must be a string URI, or a list of them"


        # Use an ATOMIC operation.  If any of the matches fail, no operation is performed
        # The "OPTIONAL MATCH" is used to compute a new initial positional value to use on the to_collection
        q = f'''
            MATCH (collection_from) , (collection_to)                
            WHERE collection_from.uri = $from_collection
              AND collection_to.uri = $to_collection            
            WITH collection_from, collection_to
                        
            // Attempt to locate Items already on the "to" Collection
            OPTIONAL MATCH (existing_ci) -[r :`{membership_rel_name}`]-> (collection_to)
            WITH r.pos AS pos, collection_from, collection_to
            WITH 
                CASE WHEN pos IS NULL THEN
                    0                           // The "to" Collection is devoid of Items
                ELSE
                    max(pos) + {cls.DELTA_POS}  // A value greater than that of the last Item
                END AS new_start_pos, collection_from, collection_to
            
            WITH range(0, size($item_list)) AS ITEM_INDEX_LIST, 
                 new_start_pos, collection_from, collection_to
            
            UNWIND ITEM_INDEX_LIST AS i         // Used to process each Collection Item in turn
                MATCH (moving_ci) -[old_r :`{membership_rel_name}`]-> (collection_from)
                WHERE moving_ci.uri = $item_list[i]
                WITH moving_ci, old_r, collection_to, 
                     new_start_pos + i * {cls.DELTA_POS} AS new_pos
                
                DELETE old_r
                MERGE (moving_ci) -[:`{membership_rel_name}` {{pos: new_pos}}]-> (collection_to)
            '''

        data_binding={"from_collection": from_collection,
                      "to_collection": to_collection,
                      "item_list": items}

        #cls.db.debug_query_print(q, data_binding=data_binding)

        status = cls.db.update_query(q, data_binding=data_binding)
        #print("bulk_relocate_to_other_collection_at_end(): status is ", status)
        # status should contain, for example {'relationships_deleted': 8, 'relationships_created': 8, 'properties_set': 8}

        number_relationships_created = status.get('relationships_created', 0)

        assert number_relationships_created <= len(items), \
            f"bulk_relocate_to_other_collection_at_end(): " \
            f"The number of created links ({number_relationships_created}) exceeds " \
            f"the number of Collection Items to move ({len(items)})"

        assert status.get('relationships_deleted', 0) == number_relationships_created, \
            f"bulk_relocate_to_other_collection_at_end(): " \
            f"The number of links deleted doesn't match " \
            f"that of the links created ({number_relationships_created})"

        assert status.get('properties_set', 0) == number_relationships_created, \
            f"bulk_relocate_to_other_collection_at_end(): " \
            f"The number of properties set doesn't match " \
            f"that of the links deleted and created ({number_relationships_created})"

        return number_relationships_created



    @classmethod
    def add_to_collection_at_end(cls, collection_uri :str, membership_rel_name: str, item_class_name: str, item_properties: dict,
                                 new_uri=None) -> str:
        """
        Create a new data node, of the class specified in item_class_name, and with the given properties -
        and add it at the end of the specified Collection, linked by the specified relationship

        EXAMPLE:  new_uri = add_to_collection_at_end(collection_id=708, membership_rel_name="BA_in_category",
                                                     item_class_name="Headers", item_properties={"text": "New Caption, at the end"}

        :param collection_uri:      The uri of a data node whose schema is an instance of the Class "Collections"
        :param membership_rel_name:
        :param item_class_name:
        :param item_properties:
        :param new_uri:             Normally, the Item ID is auto-generated, but it can also be provided (Note: MUST be unique)

        :return:                    The auto-increment "uri" assigned to the newly-created data node
        """
        #TODO:  DITCH!  The Data Node creation probably should be done ahead of time, and doesn't need to be the responsibility of this class!
        #TODO:  solve the concurrency issue - of multiple requests arriving almost simultaneously, and being handled by a non-atomic update,
        #       which can lead to incorrect values of the "pos" relationship attributes.
        #       -> Follow the new way it is handled in link_to_collection_at_end()

        assert NeoSchema.is_valid_uri(collection_uri), "The argument `collection_uri` isn't a valid URI string"
        assert type(membership_rel_name) == str, "The argument `membership_rel_name` MUST be a string"
        assert type(item_class_name) == str, "The argument `item_class_name` MUST be a string"
        assert type(item_properties) == dict, "The argument `item_properties` MUST be a dictionary"

        # TODO: this query and the one in add_data_point(), below, ought to be combined, to avoid concurrency problems
        q = f'''
            MATCH (n:BA) -[r :{membership_rel_name}]-> (c:BA {{uri: $collection_id}}) 
            RETURN max(r.pos) AS max_pos
            '''
        data_binding = {"collection_id": collection_uri}

        max_pos = cls.db.query(q, data_binding, single_cell="max_pos")  # This will be None if the Collection has no elements

        if max_pos is None:
            pos = 0                         # Arbitrary initial value for cases when the Collection has no elements
        else:
            pos = max_pos + cls.DELTA_POS   # Go some distance past the end

        data_binding = item_properties

        #cls.db.debug_print(q, data_binding, "add_to_collection_at_end", True)

        return NeoSchema.add_data_point_OLD(class_name=item_class_name,
                                            data_dict=data_binding,
                                            labels=["BA", item_class_name],
                                            connected_to_id=collection_uri, connected_to_labels="BA",
                                            rel_name=membership_rel_name,
                                            rel_prop_key="pos", rel_prop_value=pos,
                                            new_uri=new_uri
                                            )



    @classmethod
    def add_to_collection_after_element(cls, collection_uri :str, membership_rel_name: str,
                                        item_class_name: str, item_properties: dict, insert_after :str,
                                        new_uri=None) -> str:
        """
        Create a new data node, of the class specified in item_class_name, and with the given properties -
        and add to the specified Collection, linked by the specified relationship and inserted after the given collection Item
        (in the context of the positional order encoded in the relationship attribute "pos")

        :param collection_uri:      The uri of a data node whose schema is an instance of the Class "Collections"
        :param membership_rel_name: The name of the relationship to which the positions ("pos" attribute) apply
        :param item_class_name:     Name of the Class for the newly-created node
        :param item_properties:     Dictionary with the properties of the newly-created node
        :param insert_after:        The URI of the element after which we want to insert
        :param new_uri:             Normally, the Item ID is auto-generated, but it can also be provided (Note: MUST be unique)

        :return:                    The auto-increment "uri" assigned to the newly-created data node
        """
        #TODO:  the Data Node creation perhaps should be done ahead of time, and doesn't need to be the responsibility of this class!
        #TODO: solve the concurrency issue - of multiple requests arriving almost simultaneously, and being handled by a non-atomic update,
        #       which can lead to incorrect values of the "pos" relationship attributes.
        #       -> Follow the new way it is handled in link_to_collection_at_end()

        assert NeoSchema.is_valid_uri(collection_uri), "The argument `collection_uri` isn't a valid URI string"
        assert type(membership_rel_name) == str, "The argument `membership_rel_name` MUST be a string"
        assert type(item_class_name) == str, "The argument `item_class_name` MUST be a string"
        assert type(item_properties) == dict, "The argument `item_properties` MUST be a dictionary"
        assert type(insert_after) == str, "The argument `insert_after` MUST be a string"

        q = f'''
        MATCH (n_before :BA {{uri: $insert_after}})-[r_before :{membership_rel_name}] 
                    -> (c :BA {{uri: $collection_id}}) <-
                                            [r_after :{membership_rel_name}]-(n_after :BA)
        WHERE r_after.pos > r_before.pos
        RETURN r_before.pos AS pos_before, r_after.pos AS pos_after
        ORDER BY pos_after
        LIMIT 1
        '''
        #EXAMPLE:
        '''
        MATCH (n_before :BA {uri: 717})-[r_before :BA_in_category] -> (c :BA {uri: 708}) <-[r_after :BA_in_category]-(n_after :BA)
        WHERE r_after.pos > r_before.pos
        RETURN r_before.pos AS pos_before, r_after.pos AS pos_after
        ORDER BY pos_after
        LIMIT 1
        '''

        data_binding = {"collection_id": collection_uri, "insert_after": insert_after}

        # ALTERNATE WAY OF PHRASING THE QUERY:
        '''
        MATCH (n_before:BA {uri: 717})-[r_before :BA_in_category] -> (c:BA {uri: 708}) <- [r_after :BA_in_category]-(n_after :BA)
        WITH r_before.pos AS pos_before, r_after
        WHERE r_after.pos > pos_before
        RETURN pos_before, r_after.pos AS pos_after
        ORDER BY pos_after
        LIMIT 1
        '''

        result = cls.db.query(q, data_binding, single_row=True)
        print(result)
        if result is None:
            # An empty find is indicative of either an "insert at the end" (no n_after found),
            #       or a bad insert_after value that matches no node
            node = NeoSchema.search_data_node(uri = insert_after)
            if node is None:
                raise Exception(f"There is no node with the `uri` value ({insert_after}) passed by `insert_after`")

            print("It's case of insert AT THE END")
            return cls.add_to_collection_at_end(collection_uri, membership_rel_name, item_class_name, item_properties, new_uri=new_uri)


        pos_before = result["pos_before"]
        pos_after = result["pos_after"]

        if pos_after == pos_before + 1:
            # There's no room; shift everything that is past that position, by a count of DELTA_POS
            print(f"********* SHIFTING DOWN ITEMS whose `pos` value is {pos_after} and above  ***********")
            cls.shift_down(collection_id=collection_uri, membership_rel_name=membership_rel_name, first_to_move=pos_after)
            new_pos = pos_before + int(cls.DELTA_POS/2)			# This will be now be the empty halfway point
        else:
            new_pos = int((pos_before + pos_after) / 2)		    # Take the halfway point, rounded down


        return NeoSchema.add_data_point_OLD(class_name=item_class_name,
                                            data_dict=item_properties,
                                            labels=["BA", item_class_name],
                                            connected_to_id=collection_uri, connected_to_labels="BA",
                                            rel_name=membership_rel_name,
                                            rel_prop_key="pos", rel_prop_value=new_pos,
                                            new_uri=new_uri
                                            )

        #link_to = [{"labels": "BA", "key": "uri", "value": collection_id,
        #            "rel_name": membership_rel_name, "rel_attrs": {"pos": new_pos}}]

        #new_neo_id = cls.db.create_node_with_relationships(labels="BA", properties=item_properties, connections=link_to)

        #uri = NeoSchema.register_existing_data_node(class_name=item_class_name, existing_neo_id=new_neo_id, new_uri=new_uri)

        #return uri



    @classmethod
    def shift_down(cls, collection_id :str, membership_rel_name :str, first_to_move :int) -> int:
        """
        Shift down the positions (values of the "pos" relationship attributes) by an arbitrary fixed amount,
        starting with nodes with the specified position value (and all greater values);
        this operation applies to nodes linked to the specified Collection thru a relationship with the given name.

        :param collection_id:       The uri of a data node whose schema is an instance of the Class "Collections"
        :param membership_rel_name: The name of the relationship to which the positions ("pos" attribute) apply
        :param first_to_move:       All position ("pos") values greater or equal to this one will get shifted down
        :return:                    The number of modified items
        """
        # Starting with a particular Collection node, look at all its relationships whose name is specified by membership_rel_name,
        #       and increase the value of the "pos" attributes on those relationships if their current values is at least first_to_move
        q = f'''
        MATCH (c:BA {{uri: $collection_id}}) <- [r :{membership_rel_name}] - (n :BA)
        WHERE r.pos >= $first_to_move
        SET r.pos = r.pos + {cls.DELTA_POS}
        '''
        data_binding = {"collection_id": collection_id, "first_to_move": first_to_move}

        status = cls.db.update_query(q, data_binding)
        return status.get("properties_set", 0)



from brainannex.media_manager import MediaManager
from brainannex.full_text_indexing import FullTextIndexing
from brainannex.neo_schema.neo_schema import NeoSchema


class Notes:
    """
    Plugin-provided handlers for "notes"  (HTML-formatted text)
    """


    @classmethod
    def default_folder(cls):
        """
        Specify the desired name for the default folder to contain "notes" media
        """
        return "notes"



    @classmethod
    def initialize_schema(cls) -> None:
        """
        Initialize the Schema needed by this plugin

        :return:    None
        """
        pass    # TODO: follow the example in Images



    @classmethod
    def delete_content_before(cls, uri :str) -> None:
        """
        Invoked just prior to deleting the data node

        :param uri: A string with the URI ("item ID") of the Content Item
        :return:    None.  If index isn't found, an Exception is raised
        """
        #print(f"***** DELETING INDEXING for item {uri}")
        # TODO: maybe the Core can take care of this,
        #       for all Content Items that make use of word indexing
        content_id = NeoSchema.get_data_node_internal_id(uri=uri)
        FullTextIndexing.remove_indexing(content_id)



    @classmethod
    def delete_content_successful(cls, uri :str) -> None:
        pass    # No action needed



    @classmethod
    def add_content(cls, uri :str, data_binding: dict) -> dict:
        """
        Special handling for Notes (ought to be specified in its Schema):
               the "body" value is to be stored in a file named "notes-ID.htm", where ID is the uri,
               and NOT be stored in the database.  Instead, store in the database:
                       basename: "notes-ID"
                       suffix: "htm"

        :param uri:         A string with the URI of the Content Item
        :param data_binding:
        :return:            The altered data_binding dictionary.  In case of error, an Exception is raised.
        """
        # Save and ditch the "body" attribute - which is not to be stored in the database
        body = data_binding["body"]
        del data_binding["body"]

        # Create a file
        basename = f"notes-{uri}"   # TODO: change this convention
                                    #       also needs changing in notes.js
        suffix = "htm"
        filename = basename + "." + suffix
        print(f"Creating file named `{filename}`")
        #print("    File contents:")
        #print(body)
        MediaManager.save_into_file(body, filename, class_name="Notes") #TODO: turn to singular

        # Introduce new attributes, "basename" and "suffix", to be stored in the database
        data_binding["basename"] = basename
        data_binding["suffix"] = suffix

        return data_binding



    @classmethod
    def before_update_content(cls, item_data :dict) -> dict:
        """
        Invoked before a Content Item of this type gets updated in the database

        Special handling for Notes:
        the "body" value is to be stored in an HTML file,
        and NOT be stored in the database.
        What's stored in the database are the basename and suffix (extension) of the file;
        EXAMPLE:
               basename: "n-123"
               suffix: "htm"

        :param item_data:   A dict with various fields for this Note
        :return:            The altered item_data dictionary
        """
        # TODO: perhaps a special handling of a "body" for media-based Content Items
        #       could be a standard service by the Core module, based on a Schema spec
        body = item_data["body"]
        basename = item_data["basename"]

        #uri = data_binding["uri"]

        if basename == "undefined":
            raise Exception("before_update_content(): got passed the value 'undefined' "
                            "for the key `basename`")

        # Overwrite the HTML file with the body of the notes
        filename = basename + ".htm"
        print(f"Overwriting file named `{filename}`")
        #print("    File contents:")
        #print(body)
        MediaManager.save_into_file(body, filename, class_name="Notes")     # TODO: turn to singular

        revised_item_data = item_data.copy()   # Clone an independent copy

        # Ditch the "body" attribute - which is not to be stored in the database
        del revised_item_data["body"]

        return revised_item_data



    @classmethod
    def new_content_item_successful(cls, uri :str, pars: dict) -> None:
        """
        Invoked after a new Content Item of this type gets successfully added

        :param uri:     A string with the URI of the Content Item
        :param pars:    Dict with the various properties of this Content Item
        :return:        None
        """
        body = pars.get("body")

        # TODO: do in separate thread or process
        # TODO: maybe the Core can take care of this,
        #       for all Content Items that make use of word indexing
        unique_words = FullTextIndexing.extract_unique_good_words(body)
        content_id = NeoSchema.get_data_node_internal_id(uri=uri)
        n_words = len(unique_words)
        print(f"new_content_item_successful(): CREATING INDEXING for item `{uri}`. "
              f"Found {n_words} unique words; first few: {list(unique_words)[:10]}")
        FullTextIndexing.new_indexing(internal_id=content_id, unique_words=unique_words)



    @classmethod
    def update_content_item_successful(cls, uri :str, pars :dict) -> None:
        """
        Invoked after a Content Item of this type gets successfully updated in the database

        :param uri:     A string with the URI of the Content Item
        :param pars:    Dict with the various properties of this Content Item
        :return:        None
        """
        body = pars.get("body")

        # TODO: do in separate thread or process
        # TODO: maybe the Core can take care of this,
        #       for all Content Items that make use of word indexing
        unique_words = FullTextIndexing.extract_unique_good_words(body)
        content_id = NeoSchema.get_data_node_internal_id(uri=uri)
        n_words = len(unique_words)
        print(f"update_content_item_successful(): UPDATING INDEXING for item `{uri}`. "
              f"Found {n_words} unique words; first few: {list(unique_words)[:10]}")
        FullTextIndexing.update_indexing(content_uri=content_id, unique_words=unique_words)



from brainannex.media_manager import MediaManager
from brainannex.neo_schema.neo_schema import NeoSchema


class Images:
    """
    Plugin-provided custom interface for "images"
    """

    SCHEMA_CLASS_NAME = "Images"   # TODO: turn to singular



    @classmethod
    def default_folder(cls):
        """
        Specify the desired name for the default folder to contain image media
        """
        return "images"



    @classmethod
    def initialize_schema(cls) -> None:
        """
        Initialize the Schema needed by this plugin

        :return:    None
        """
        assert NeoSchema.is_valid_class_name(cls.SCHEMA_CLASS_NAME), \
            f"initialize_schema(): attempting to create a Schema Class with an invalid name: '{cls.SCHEMA_CLASS_NAME}'"

        if not NeoSchema.class_name_exists("Media"):
            NeoSchema.create_class_with_properties(name="Media", strict=True,
                                               properties=["basename", "suffix"])

        NeoSchema.create_class_with_properties(name=cls.SCHEMA_CLASS_NAME, strict=True, code="i",
                                               properties=["width", "height", "caption", "date_created"],
                                               class_to_link_to="Media", link_name="INSTANCE_OF", link_dir="OUT")

        if not NeoSchema.class_name_exists("Directory"):
            NeoSchema.create_class_with_properties(name="Directory", strict=True,
                                                   properties=["name", "description"])
            NeoSchema.create_class_relationship(from_class="Media", to_class="Directory", rel_name="BA_stored_in")



import re                               # For REGEX
import pandas as pd



"""
    MIT License.  Copyright (c) 2021-2024 Julian A. West and the BrainAnnex.org project
"""



class DocumentationGenerator:
    """
    To generate the HTML for a documentation page, from a python file
    (for best results, it's expected to be a file following some styling conventions
    as done in the BrainAnnex project)
    """


    @classmethod
    def import_python_file(cls, basename, full_filename) -> str:
        """
        Parse a python source file (expected to conform to a few conventions)
        and generate HTML that can be used to create a documentation page

        :param basename:        EXAMPLE: "my_file_being_uploaded.txt"
        :param full_filename:   EXAMPLE: "D:/tmp/my_file_being_uploaded.txt"
        :return:                HTML code to populate a documentation page
        """
        # Read in the python file
        n_chars_to_show = 400
        try:
            with open(full_filename, 'r') as fh:
                file_contents = fh.read()
                print(f"\n--- First {n_chars_to_show} bytes of uploaded file:\n{file_contents[:n_chars_to_show]}")
        except Exception:
            return f"import_python_file(): File I/O failed (on uploaded file {full_filename})"


        pattern = cls.define_pattern()      # String that will be used for REGEX parsing
        print("Pattern used for the matches: ", pattern)

        all_matches = re.findall(pattern, file_contents, re.DOTALL)
        # It returns (possibly-empty) list of tuples
        # OR a list of strings (if there's only 1 capture group in pattern)
        print(f"\n-- {len(all_matches)} MATCHES --")
        for m in all_matches:
            abridged_match = []
            for el in m:
                abridged_match.append(el[:60])
            print(f"match: {abridged_match}")


        if all_matches:     # If the list is not empty, i.e. if matches were found
            #print(f"{len(all_matches)} MATCH(ES) found")
            # Produce a simple table to present the various matches that were found
            scan_results = "<table border='1' style='border-collapse: collapse'>"
            for match_instance in all_matches:   # Consider each match in turn
                #print("Overall Single Match: " , match_instance) # This would normally be a tuple of capture groups
                # (which we previously turned to list, with 2 field added)
                scan_results += "<tr>"
                for item in match_instance:
                    scan_results += f"<td>{item}</td>"
                scan_results += "</tr>"

            scan_results += f"</table>"
        else:
            print("NO MATCHES found")
            return f"File `{basename}` uploaded successfully, but <b>NO MATCHES</b> found"


        # Put together the parsing data as a Pandas dataframe
        column_names = ["method_name", "args", "return_value", "comments", "class_name", "class_description"]
        df = pd.DataFrame(all_matches, columns = column_names)
        #print("Number of records matched:", df.count())
        '''
        print(df.head(10))
        print("...")
        print(df.tail(10))
        '''

        # Modify the "args" column, by applying the "clean_up_args" to each of its entries
        df["args"] = df["args"].map(cls.clean_up_args)


        # Produce HTML code to documented the python file from its parsing data
        htm = cls.generate_documentation(df)

        safe_htm = htm.replace("<", "&lt;").replace(">", "&gt;")

        # TODO:  < , > and & in code comments are still not sufficiently protected
        #        Need to replace  <  with  &lt;
        #        Need to replace  &  with  &amp;
        #        Try it out on documenting the method FullTextIndexing.split_into_words()

        return f"File `{basename}` uploaded successfully.  <b>{len(all_matches)} MATCH(ES)</b> found.  Nothing added to database.  " \
               f"Scan results:<br><br>{scan_results}<br><br>" \
               f"<b>HTML:</b><br><br><pre>{safe_htm}</pre>"



    @classmethod
    def define_pattern(cls) -> str:
        """
        Pattern for the creation of documentation from python files.

        The python files has some expectations about their formatting; for example, as used in neo_schema.py

        Compose and return a REGEX pattern for parsing of data files, for use in import_datafile()

        The pattern is expected to be used in a re.findall() that uses re.DOTALL as the last argument

        :return:    A string with a REGEX pattern
        """
        # The R before the string escapes all characters ("raw strings" aka "verbatim strings")

        '''
        PART A - Python Class Methods
        '''

        pattern_1 = R'def\s+([a-zA-Z_][a-zA-Z0-9_]*)'   # Match and capture the method name:
                                                        #   "def" followed by 1 or more blanks, followed by
                                                        #   (one letter or underscore, followed by any number of letters, numbers or underscores)

        pattern_2 = R'\((.*?)\)'                        # Match and capture (non-greedy) everything inside the round parentheses after method name

        #pattern_3 = R'(?:\s*->\s*(.*?)\s*)?:'
        #pattern_3 = R'(?:\s*->\s*(.*?)\s*)?:(?:\s*#.*?\n)?'
        pattern_3 = R'(?:\s*->\s*(.*?)\s*)?:.*?\n'              # Match and capture (non-greedy) the method's return type - which may or may not be present
                                                                # Advance (there might be blanks and/or optional comments) until the end of the line
        '''
            (?:                 Start of NON-capturing grouping
                \s*                 0+ blanks
                ->                  Literal "->"
                \s*                 0+ blanks
                (                   Start of capture group
                    +*?                 Any single character, 1 or more times (non-greedy)
                )                   End of capture group
                \s*                 0+ blanks
            )                   End of non-capturing grouping
            ?                   Make the preceding group optional
            :                   Literal ":"
            .*?                 Any single character, 0 or more times (non-greedy)
            \n                  End of line           
        '''


        pattern_4 = R'(?:\s+"""(.*?)""")?'          # Match and capture (non-greedy) everything within the following pair of """
        '''
            (?:                 Start of NON-capturing grouping
                \s+                 1+ blanks
                """                 Literal triple double quotes
                (                   Start of capture group
                    .*?                 Any single character, 0 or more times (non-greedy)
                                        Note: newlines also matched because the calling functions uses re.DOTALL
                                              as the 3rd argument in re.findall()
                )                   End of capture group
                """                 Literal triple double quotes
            )                   End of non-capturing grouping
            ?                   Make the preceding group optional
        '''


        pattern_A = pattern_1 + pattern_2 + pattern_3 + pattern_4


        '''
        PART B - Python Class Names
        '''
        # Match and capture a python class name
        #       EXAMPLE 1:  "class NeoAccessCore:"
        #       EXAMPLE 2:  "class NeoAccess(NeoAccessCore):"
        pattern_1 = R'class\s+([a-zA-Z][a-zA-Z0-9_]*)(?:\([a-zA-Z][a-zA-Z0-9_]*\))?\s*:'
        '''
            class               Literal "class"
            \s+                 1+ blanks
            (                   Capture start
                [a-zA-Z]            letter
                [a-zA-Z0-9_]*       0+ alphanumeric or underscore
            )                   Capture end
            (?:                 Start of NON-capturing grouping
                \(                  Literal "("
                [a-zA-Z]            letter
                [a-zA-Z0-9_]*       0+ alphanumeric or underscore  
                \)                  Literal ")"    
            )                   End of non-capturing grouping
            ?                   Make the preceding group optional
            \s*                 0+ blanks
            :                   Literal ":"
        '''

        pattern_2 = R'.+?"""(.*?)"""'               # Match and capture (non-greedy) everything within the following pair of """
        #   The .+? at the beginning = 1 or more characters (non-greedy).  Note: that required character is the preceding newline
        pattern_B = pattern_1 + pattern_2


        pattern = f"(?:{pattern_A})|(?:{pattern_B})"    # Deal with alternations.  Note: "?:" means that the parentheses are NOT a capture group

        return pattern



    @classmethod
    def clean_up_args(cls, s:str) -> str:
        """
        Strip off the leading "cls" or "self", together with the comma after it,
        from a string that is meant to represent the arguments of a python class method declaration.
        Leading/trailing blanks are also eliminated

        EXAMPLES:   "   cls, r, x    "                  gives  "r, x"
                    "   self   , r:int,  x :list    "   gives  "r:int,  x :list"

        :param s:   String with a list of arguments of a python class method
        :return:    Cleanup-up version, stripped of leading "cls" or "self"
        """
        #TODO: also manage cases where s is simply "cls" or "self" (i.e, no args)

        pos_first_comma = s.find(",")       # Position in the string of the first comma
        if pos_first_comma == -1:           # If not found
            return s.strip()

        first_arg = s[:pos_first_comma]
        first_arg = first_arg.strip()       # Cleaned-up version of the 1st argument (the expected "cls" or "self")

        if (first_arg != "cls") and (first_arg != "self"):      # If not found
            return s.strip()

        abridged = s[pos_first_comma+1 :]   # Ditch the early part of the string, up to - and including - the first comma
        return abridged.strip()             # Eliminate lading/trailing blanks



    @classmethod
    def generate_documentation(cls, df :pd.DataFrame) -> str:
        """
        Print out, and return as a string, the HTML code to create a documentation page,
        from a Pandas data frame containing the data about the various elements
        of the python file.
        TODO: probably switch to a Flask template

        Note: the HTML code also contains references to some CSS classes for styling.

        :param df:  A Pandas data frame, with the following columns:
                        class_name, class_description, method_name, args, return_value, comments
        :return:    A string with HTML code
        """

        toc = "<div class='sidebar-left'>\n"    # Table of contents, with links to Classes and methods
        htm = ""                                # HTML for the main body of the documentation page, EXCLUSIVE of the above Table of Contents
        python_class_name = ""
        first_python_class = True               # Flag indicating whether the currently-parsed python Class is the first being encountered

        for ind in df.index:    # EXAMPLE of df.index: RangeIndex(start=0, stop=11, step=1)
            # For each row in the Pandas data frame

            if df['class_name'][ind]:                   # Start documenting a new pythons class
                python_class_name = df['class_name'][ind]
                python_class_description = df['class_description'][ind]

                toc += f"\n    <br><a href='#{python_class_name}' style='font-weight:bold; font-size:18px'>Class {python_class_name}</a><br>\n"

                if first_python_class:
                    first_python_class = False
                else:
                    htm += "<br><br><hr>"       # Skipped if it's the first Class

                htm += f"<a name='{python_class_name}'></a>\n"
                htm += f"<h1 class='class-name'>Class {python_class_name}</h1>\n"
                htm += f"<pre>{python_class_description}</pre>\n\n\n"

            elif "____" in df['method_name'][ind]:      # A BrainAnnex styling convention to indicate a new section
                section_name = df['method_name'][ind]
                clean_name = section_name.replace("_", " ").strip()

                toc += f"\n    <span style='margin-left:15px; font-weight:bold'>{clean_name}:</span><br>\n"
                htm += f"<br><h2 class='section-header'>{clean_name}</h2>\n\n"

            else:           # Document an individual class method
                method_name = df['method_name'][ind]
                anchor_name = f"{python_class_name}_{method_name}"   # Needed because some method names (such as __init__) may appear in multiple classes
                                                                     # EXAMPLE:  "NeoAccess_query"

                toc += f"    <a href='#{anchor_name}' style='margin-left:40px'>{method_name}</a><br>\n"

                htm += f"<a name='{anchor_name}'></a>\n"
                htm += "<table class='cd-main'>\n"
                htm += "<tr><th>name</th><th>arguments</th><th>returns</th></tr>\n"

                htm += "<tr>"
                #print(df["method_name"][ind], df["args"][ind], df["comments"][ind])
                htm += f"<td class='cd-fun-name'>{df['method_name'][ind]}</td>"
                htm += f"<td>{df['args'][ind]}</td>"
                htm += f"<td>{df['return_value'][ind]}</td>"

                htm += "</tr>\n"

                htm += "<tr>"
                htm += "<td colspan=3 class='cd-description'>\n"
                htm += f"<pre>{df['comments'][ind]}</pre>\n"
                htm += "</td>\n"
                htm += "</tr>\n"
                htm += "</table>\n\n\n"


        toc += "\n<br>\n</div>    <!-- sidebar-left -->\n\n\n\n\n"     # Complete the Table of Contents

        '''
        print("###################################################################################")
        print(toc)
        print(htm)
        print("###################################################################################")
        '''

        return toc + htm



from typing import Union, List
from brainannex.neo_schema.neo_schema import NeoSchema
from brainannex.collections import Collections
from neoaccess import NeoAccess
import pandas as pd


class Categories:
    """
    Library for Category-related operations.

    An entity to which a variety of nodes (e.g. representing records or media)
    is attached, with a positional attribute.

    Categories also have "subcategory" and "see_also" relationships with other categories.
    """

    db = None   # MUST be set before using this class!
                # Database-interface object is a CLASS variable, accessible as cls.db

    DELTA_POS = 20      # Arbitrary shift in "pos" value; best to be even, and not too small nor too large



    @classmethod
    def set_database(cls, db :NeoAccess) -> None:
        """
        IMPORTANT: this method MUST be called before using this class!

        :param db:  Database-interface object, created with the NeoAccess library
        :return:    None
        """

        assert type(db) == NeoAccess, \
            "Categories.set_database(): argument passed isn't a valid NeoAccess object"

        cls.db = db     # Save the database object

        # Initialize dependencies
        NeoSchema.set_database(db)
        Collections.set_database(db)



    @classmethod
    def initialize_categories(cls) -> (int, str):
        """
        Create a new Schema Class node that represents "Categories",
        and make it an "INSTANCE_OF" of the "Collections" Class

        :return:    An (int, str) pair of integers with the internal database ID
                        and the unique uri assigned to the new Class node
        """
        (int_dbase_id, uri) = NeoSchema.create_class_with_properties(name="Categories",
                                                        properties=["name", "remarks", "uri", "root"],
                                                        strict=True)

        NeoSchema.create_class_relationship(from_class="Categories", to_class="Categories",
                                            rel_name="BA_subcategory_of", use_link_node=False)

        NeoSchema.create_class_relationship(from_class="Categories", to_class="Categories",
                                            rel_name="BA_see_also", use_link_node=False)

        if not NeoSchema.class_name_exists("Collections"):
            Collections.initialize_collections()
            NeoSchema.create_class_relationship(from_class="Categories", to_class="Collections",
                                                rel_name="INSTANCE_OF", use_link_node=False)

        return (int_dbase_id, uri)





    #####################################################################################################

    '''                                 ~   LOOKUP CATEGORY DATA   ~                                  '''

    def ________LOOKUP_CATEGORIES________(DIVIDER):
        pass        # Used to get a better structure view in IDEs
    #####################################################################################################

    @classmethod
    def get_category_info(cls, category_uri :str) -> dict:
        """
        Return the Name and Remarks field values attached to the given Category

        :param category_uri:    A string identifying the desired Category
        :return:                The Category's name (or a blank dictionary if not found)
                                    EXAMPLE:  {"uri": "123", "name": "Astronomy", "remarks": "except cosmology"}
        """
        # Note: since the category_uri is a primary key,
        #       specifying a value for the labels and the "schema_code" property is for redundancy
        #return NeoSchema.search_data_node(uri=category_uri, labels="BA", properties={"schema_code": "cat"})
        return NeoSchema.get_data_node(class_name="Categories", node_id=category_uri, id_key="uri")



    @classmethod
    def is_root_category(cls, category_uri :str) -> bool:
        """
        Return True if the given ID corresponds to the ROOT Category, or False otherwise

        :param category_uri:    A string identifying the desired Category
        :return:                True if the given ID corresponds to the ROOT Category, or False otherwise
        """
        assert NeoSchema.is_valid_uri(category_uri), \
            "is_root_category(): the argument `category_uri` is not a valid URI string"

        # NOTE: historically, "1" has been used for the ROOT Category; however, now that uri is shared
        #       among all types of plugins, maybe a different approach would be better (such as an attribute in the node)
        return True if category_uri == "1" else False    # TODO: this will eventually have to be managed differently



    @classmethod
    def get_all_categories(cls, exclude_root=True, include_remarks=False) -> [dict]:
        """
        Return all the existing Categories - possibly except the root -
        as a list of dictionaries with keys 'uri', 'name', 'pinned' and, optionally, 'remarks',
        sorted by name.

        :param exclude_root:    If True, the root Category is omitted
        :param include_remarks: If True, the 'remarks' property is included alongside all others

        :return:    A list of dictionaries.  EXAMPLE:
                        [{'uri': '2', 'name': 'Work', 'remarks': 'Current or past'},
                         {'uri': '3', 'name': 'Hobbies', pinned: True} ]
        """
        clause = ""
        if exclude_root:
            clause = "WHERE (cat.root <> true OR cat.root is NULL)"


        remarks_subquery = ", cat.remarks AS remarks"  if include_remarks else ""

        # TODO: switch to using the Schema library datanode operations
        q =  f'''
             MATCH (cat:Categories)-[:SCHEMA]->(:CLASS {{name:"Categories"}}) 
             {clause}
             RETURN cat.uri AS uri, cat.name AS name, cat.pinned AS pinned {remarks_subquery}
             ORDER BY toLower(cat.name)
             '''

        # Notes: Sorting must be done across names of consistent capitalization, or "GSK" will appear before "German"!

        result =  cls.db.query(q)

        # Ditch all the MISSING "pinned" values  (TODO: let the Schema layer handle this!)
        for item in result:
            if item["pinned"] is None:
                del item["pinned"]     # To avoid a dictionary entry of the type  'pinned': None

        # If remarks are being included, ditch all the MISSING "remarks" values
        if include_remarks:
            for item in result:
                if item["remarks"] is None:
                    del item["remarks"]     # To avoid a dictionary entry of the type  'remarks': None

        return result



    @classmethod
    def count_subcategories(cls, category_uri :str) -> int:
        """
        Return the number of (direct) Subcategories of the given Category
        TODO: maybe turn into a method of NeoSchema :  count_inbound_rels(labels="BA, uri=category_uri, class_name="Categories")

        :param category_uri:A string identifying the desired Category
        :return:            The number of (direct) Subcategories of the given Category; possibly, zero
        """

        match = cls.db.match(labels="Categories",
                             properties={"uri": category_uri})

        return cls.db.count_links(match, rel_name="BA_subcategory_of", rel_dir="IN")


    @classmethod
    def count_parent_categories(cls, category_uri :str) -> int:
        """
        Return the number of (direct) Subcategories of the given Category
        TODO: maybe turn into a method of NeoSchema :  count_outbound_rels(labels="BA, uri=category_uri, class_name="Categories")

        :param category_uri:A string identifying the desired Category
        :return:            The number of (direct) parent categories of the given Category; possibly, zero
        """

        match = cls.db.match(labels="Categories",
                             properties={"uri": category_uri})

        return cls.db.count_links(match, rel_name="BA_subcategory_of", rel_dir="OUT")



    @classmethod
    def get_subcategories(cls, category_uri :str) -> [dict]:
        """
        Return all the (immediate) subcategories of the given category,
        as a list of dictionaries with all the keys of the Category Class
        EXAMPLE:
            [{'uri': '2', 'name': 'Work', remarks: 'outside employment'},
             {'uri': '3', 'name': 'Hobbies'}]

        :param category_uri:A string identifying the desired Category
        :return:            A list of dictionaries
        """
        match = cls.db.match(labels="Categories", key_name="uri", key_value=category_uri)

        return cls.db.follow_links(match, rel_name="BA_subcategory_of", rel_dir="IN",
                                   neighbor_labels="Categories")



    @classmethod
    def get_parent_categories(cls, category_uri :str) -> [dict]:
        """
        Return all the (immediate) parent categories of the given Category,
        as a list of dictionaries with all the keys of the Category Class
        EXAMPLE:
            [{'uri': '2', 'name': 'Work', remarks: 'outside employment'},
             {'uri': '3', 'name': 'Hobbies'}]

        :param category_uri:A string identifying the desired Category
        :return:            A list of dictionaries
        """
        match = cls.db.match(labels="Categories",
                             properties={"uri": category_uri})

        return cls.db.follow_links(match, rel_name="BA_subcategory_of", rel_dir="OUT",
                                   neighbor_labels="Categories")



    @classmethod
    def get_sibling_categories(cls, category_internal_id: int) -> [dict]:
        """
        Return the data of all the "siblings" nodes of the given Category

        :param category_internal_id:    The internal database ID of a "Category" data node
        :return:                        A list of dictionaries, with one element for each "sibling";
                                            each element contains the 'internal_id' and 'neo4j_labels' keys,
                                            plus whatever attributes are stored on that node.
                                            EXAMPLE of single element:
                                            {'name': 'French', 'internal_id': 123, 'neo4j_labels': ['Categories', 'BA']}
        """

        #TODO: switch to this after the next update of NeoAccess
        #result = cls.db.get_siblings(internal_id=category_internal_id, rel_name="BA_subcategory_of", order_by="name")

        q = f"""
                MATCH (n) - [:BA_subcategory_of] -> (parent) <- [:BA_subcategory_of] - (sibling)
                WHERE id(n) = {category_internal_id}
                RETURN DISTINCT sibling
                ORDER BY toLower(sibling.name)
            """
        result = cls.db.query_extended(q, flatten=True)

        # Ditch unneeded attributes
        #for item in result:
        #    del item["neo4j_labels"]

        return result



    @classmethod
    def create_parent_map(cls, category_uri :str) -> dict:
        """
        Consider the set comprising the given Category and all its ancestors (i.e. all its super-categories),
        up to a maximum hop length.

        Create and return a dictionary that maps each of the uri in that set of Categories,
        to a list of the uri's of its parent Categories.

        :param category_uri:A string identifying the desired Category
        :return:            A dictionary mapping integers into lists of integers.
                            The keys are uri's of the given Category and any of its ancestors (super-categories),
                            up to a maximum hop length;
                            the values are lists of the uri's of the parent categories of the Category specified by the key
                            EXAMPLES:       {'123': ['1']}                                  # The given category (123) is a child of the root (1)
                                            {'823': ['709'], '709': ['544'], '544': ['1']}  # A simple 3-hop path from Category 823 to the root (1) :
                                                                                            #       823 is subcategory of 709,
                                                                                            #       which is subcategory of 544, which is subcategory of the root
                                            {'814': ['20', '30'], '20': ['1'], '30': ['79'], '79': ['1']}
                                                                                            # Two paths from Category 814 to the root;
                                                                                            #       814 is subcategory of 20 and 30;
                                                                                            #       20 is an subcategory of the root,
                                                                                            #       while with 30 we have to go thru an extra hop
        """

        # Based on the "BA_subcategory_of" relationship,
        #       extract a set of maps of the form child (c) -> all its parent categories,
        #       where the child c is any ancestor node of the given Category node.
        #       A limit is imposed on the max length of the path
        q = '''
            MATCH (start :BA:Categories {uri:$category_uri})-[:BA_subcategory_of*0..9]->
                  (c :Categories)-[:BA_subcategory_of]->(p :Categories)
            WITH c, collect(DISTINCT p.uri) AS all_parents
            RETURN c.uri AS uri, all_parents
            '''
        result = cls.db.query(q, {"category_uri": category_uri})      # A list of dictionaries
        # EXAMPLE:  [{'uri': 823, 'all_parents': [709]},
        #            {'uri': 709, 'all_parents': [544]},
        #            {'uri': 544, 'all_parents': [1]}]

        parent_map = {}
        for entry in result:
            key = entry["uri"]
            value = entry["all_parents"]
            parent_map[key] = value

        # EXAMPLE of parent_map:   {823: [709], 709: [544], 544: [1]}
        return parent_map



    @classmethod
    def create_bread_crumbs(cls, category_uri :str) -> list:
        """
        Return a list of Category ID's together with token strings, providing directives for the HTML structure of
        the bread crumbs

        :param category_uri:A string with the URI of the Category whose "ancestry bread crumbs" we want to construct
        :return:            A list of Category URI's together with token strings,
                            providing directives for the HTML structure of the bread crumbs
                            EXAMPLE 1:  ['1']
                            EXAMPLE 2:  ['START_CONTAINER', ['1', 'ARROW', '799', 'ARROW', '876'], 'END_CONTAINER']
                            EXAMPLE 3:
                                [
                                    'START_CONTAINER',
                                    ['START_BLOCK',
                                                    'START_LINE', ['1', 'ARROW', '799', 'ARROW', '526'], 'END_LINE', 'CLEAR_RIGHT',
                                                    'START_LINE', ['1', 'ARROW', '61'], 'END_LINE',
                                     'END_BLOCK', 'ARROW', '814'],
                                    'END_CONTAINER'
                                ]
        """
        if category_uri == "1":    # If it is the root
            return ["1"]

        # If we get here, we're NOT at the root.

        parents_map = cls.create_parent_map(category_uri)
        #print("In create_bread_crumbs().  parents_map: ", parents_map)

        # Put together a block (to be turned into an HTML element by the front end) depicting all possible
        # breadcrumb paths from the ROOT to the current category
        return ["START_CONTAINER", cls.recursive(category_uri, parents_map) , "END_CONTAINER"]



    @classmethod
    def recursive(cls, category_uri :str, parents_map :dict) -> list:
        """

        :param category_uri:A string identifying the desired Category
        :param parents_map: The dict structure returned by create_parent_map()
        :return:
        """
        if cls.is_root_category(category_uri):
            return ["1"]    # TODO: this will eventually have to be managed differently

        parent_list = parents_map.get(category_uri, [])

        if len(parent_list) == 0:
            return []    # TODO: generate warning
        elif len(parent_list) == 1:     # If just one parent
            parent_id = parent_list[0]
            bc = cls.recursive(parent_id, parents_map)
            bc.append("ARROW")
            bc.append(category_uri)
            return bc
        else:                           # If multiple parents
            bc = ["START_BLOCK"]
            for parent_id in parent_list:
                bc.append("START_LINE")
                bc.append(cls.recursive(parent_id, parents_map))
                bc.append("END_LINE")
                if parent_id != parent_list[-1]:    # Skip if we're dealing with last element
                    bc.append("CLEAR_RIGHT")

            bc.append("END_BLOCK")
            bc.append("ARROW")
            bc.append(category_uri)
            return bc





    #####################################################################################################

    '''                                 ~   MODIFY CATEGORY DATA   ~                                  '''

    def ________MODIFY_CATEGORIES________(DIVIDER):
        pass        # Used to get a better structure view in IDEs
    #####################################################################################################


    @classmethod
    def create_categories_root(cls, data_dict=None) -> (int, str):
        """
        Create a ROOT Category node;
        and return its internal database ID and its URI

        :param data_dict:   (OPTIONAL) Dict to specify alternate desired values
                                for the "name" and "remarks" fields of the Root Category
                                (by default, "HOME" and "top level", respectively)
        :return:            The pair (internal database ID, string URI)
                                of the new Data Node just created
        """
        if data_dict is None:
            data_dict = {"name": "HOME", "remarks": "top level"}

        data_dict["root"] = True        # Flag to mark this node as the root of the Category graph

        #new_uri = NeoSchema.reserve_next_uri(namespace="Categories", prefix="cat-")
        new_uri = "cat-root"    # use a special URI

        internal_id = NeoSchema.create_data_node(class_node="Categories",
                                                 properties = data_dict,
                                                 new_uri=new_uri)
        return (internal_id, new_uri)



    @classmethod
    def add_subcategory(cls, data_dict :dict, category_uri=None) -> str:
        """
        Add a new Subcategory to a given, already existing, Category

        :param data_dict:   Dictionary with the following keys:
                                category_uri            URI to identify the Category
                                                            to which to add the new Subcategory
                                subcategory_name        The name to give to the new Subcategory
                                subcategory_remarks     (OPTIONAL)  A comment field for the new Subcategory

        :param category_uri: Optional way to specify the parent Category

        :return:            A string with the "uri" of the Category node just created
                                (which makes use of an auto-increment value)
        """
        # TODO: complete switch to named arguments rather than the archaic data_dict

        # TODO: block the addition of multiple subcategories with the same name (i.e., prevent
        #       a Category to have multiple "children" with the same name)
        if not category_uri:
            category_uri = data_dict.get("category_uri")
            assert category_uri is not None, \
                        "add_subcategory(): key `category_uri` in argument `data_dict` is missing"

        assert NeoSchema.is_valid_uri(category_uri), \
                    f"add_subcategory(): invalid category uri ({category_uri})"

        subcategory_name = data_dict.get("subcategory_name")

        #TODO: eliminate leading/trailing blanks

        if not subcategory_name:
            raise Exception(f"add_subcategory(): subcategory_name is missing")

        subcategory_remarks = data_dict.get("subcategory_remarks")  # This field is optional

        data_dict = {"name": subcategory_name}
        if subcategory_remarks:
            data_dict["remarks"] = subcategory_remarks

        parent_category_internal_id = NeoSchema.get_data_node_id(key_value=category_uri, key_name="uri")

        new_internal_id = NeoSchema.add_data_node_with_links(
                                class_name = "Categories",
                                properties = data_dict, labels = ["BA", "Categories"],
                                links = [{"internal_id": parent_category_internal_id, "rel_name": "BA_subcategory_of"}],
                                assign_uri=True)

        new_data_point = NeoSchema.search_data_node(internal_id = new_internal_id)
        assert new_data_point is not None, \
            "add_subcategory(): failure to fetch data node for the newly created subcategory"

        return new_data_point["uri"]



    @classmethod
    def delete_category(cls, uri :str) -> None:
        """
        Delete the specified Category, provided that there are no Content Items linked to it.
        In case of error or failure, an Exception is raised.

        :param uri: The uri identifying the desired Category
        :return:    None
        """
        category_uri = uri

        if cls.is_root_category(category_uri):
            raise Exception("Cannot delete the Root node")       # TODO: this will eventually have to be managed differently

        # First, make sure that there are no Content Items linked to this Category
        number_items_attached = Collections.collection_size(collection_id=category_uri, membership_rel_name="BA_in_category")

        if number_items_attached > 0:
            raise Exception(f"Cannot delete the requested Category (URI '{category_uri}') because "
                            f"it has Content Items attached to it: {number_items_attached} item(s). "
                            f"You need to first untag or delete all Items associated to it")

        if cls.count_subcategories(category_uri) > 0:
            raise Exception(f"Cannot delete the requested Category (URI '{category_uri}') because it has sub-categories. Use the Category manager to first sever those relationships")

        number_deleted = NeoSchema.delete_data_point(uri=category_uri, labels="BA")

        if number_deleted != 1:
            raise Exception(f"Failed to delete the requested Category (URI '{category_uri}')")



    @classmethod
    def add_subcategory_relationship(cls, data_dict :dict) -> None:
        """
        Add a sub-category ("BA_subcategory_of") relationship
        between the specified 2 existing Categories.
        If the requested new relationship cannot be created (for example, if it already exists),
        raise an Exception

        :param data_dict:   Two keys are expected:
                                "sub"         URI to identify an existing Category node
                                                that is to be made a sub-category of another one
                                "cat"         URI to identify an existing Category node
                                                that is to be made the parent of the other Category

        :return:            None.  If the requested new relationship could not be created,
                                raise an Exception
        """

        subcategory_uri = data_dict["sub"]
        category_uri = data_dict["cat"]


        # Notice that, because the relationship is called a SUB-category, the subcategory is the "parent"
        #   (the originator) of the relationship
        try:
            NeoSchema.add_data_relationship(from_id=subcategory_uri, to_id=category_uri,
                                            rel_name="BA_subcategory_of", id_type="uri")
        except Exception as ex:
            raise Exception(f"add_subcategory_relationship(): Unable to create a subcategory relationship. {ex}")




    """
    NOTE: the next 2 methods, below, may be the future prototype of plugin-specific methods...
          Nothing is returned if all is good, but an Exception is raised in case of problems.
          The methods only handle the plugin-specific part; the "main action" is done by the core method,
          AFTER calling this method (if provided)
    """

    @classmethod
    def add_relationship_before(cls, from_id :str, to_id :str,
                                rel_name :str) -> None:
        """
        A handler to be invoked by the core module before a relationship involving Categories is called.

        If any restriction would apply to adding the parent/child relationship between the specified categories,
        raise an Exception.

        IMPORTANT: NO RELATIONSHIP IS ACTUALLY ADDED

        The restriction are:
            1) the subcategory node cannot be the Root Category
            2) a category cannot be a subcategory of itself

        NOTE: the "BA_subcategory_of" relationship goes FROM the subcategory TO the parent category node

        :param from_id:     String with the uri of the subcategory node
        :param to_id:       String with the uri of the parent-category node
        :param rel_name:    NOT USED
        :return:            None.  If the requested new relationship should not be created, raise an Exception
        """
        # If the sub-category is the Root Category, raise an Exception
        if cls.is_root_category(from_id):        # TODO: this will eventually have to be managed differently
            raise Exception("Cannot add the relationship because the Root Category cannot be made a subcategory of something else")

        # If the parent and the child are the same, raise an Exception
        assert from_id != to_id, \
            "Cannot add a relationship from a Category to itself"



    @classmethod
    def remove_relationship_before(cls, from_id: str, to_id :str,
                                   rel_name: str) -> None:
        """
        A handler to be invoked by the core module before a relationship involving Categories is called.

        If any restriction would apply to removing the parent/child relationship between the specified categories,
        raise an Exception.

        IMPORTANT: NO RELATIONSHIP IS ACTUALLY REMOVED

        The restriction is:
            *) the subcategory node cannot become orphaned as a result of the deletion

        NOTE: the "BA_subcategory_of" relationship goes FROM the subcategory TO the parent category node

        :param from_id:     String with the uri of the subcategory node
        :param to_id:       NOT USED.  String with the uri of the parent-category node
        :param rel_name:    NOT USED
        :return:            None.  If the requested new relationship should not be deleted, raise an Exception
        """
        # If the sub-category has only one parent, raise an Exception
        #print(f"In Category.remove_relationship(). from_id = {from_id}  Parent categories : {cls.get_parent_categories(from_id)}")
        assert len(cls.get_parent_categories(from_id)) != 1, \
            "Cannot sever the relationship because that would leave " \
            "the sub-category orphaned (i.e. with no parent categories)"




    @classmethod
    def switch_parent_category_relationship(cls, child_id :str, old_parent_id :str, new_parent_id :str) -> None:
        """
        TODO: not yet implemented

        Switch a parent/child relationship between the specified categories.
        Take the child away from the old parent, and re-assign to the new one.

        :param child_id:
        :param old_parent_id:
        :param new_parent_id:
        :return:
        """
        #TODO: make use of cls.db.reattach_node()
        pass    # switchParentNode($childID, $oldParentID, $newParentID)


    @classmethod
    def switch_subcategory_relationship(cls, parent_id :str, old_child_id :str, new_child_id :str) -> None:
        """
        TODO: not yet implemented

        Switch a parent/child relationship between the specified categories.
        From the parent, take the old child away , and replace it with the new child.

        :param parent_id:
        :param old_child_id:
        :param new_child_id:
        :return:
        """
        #TODO: make use of cls.db.reattach_node()
        pass    # switchChildNode($parentID, $oldChildID, $newChildID)



    @classmethod
    def pin_category(cls, uri, op :str) -> None:
        """
        Set or unset the "pinned" property of the specified Category

        :param uri: The URI of a data node representing a Category
        :param op:  Either "set" or "unset"
        :return:    None
        """
        # TODO: verify that the node is indeed a Category - or make sure that the Schema is enforced
        #       Maybe first locate the data node by multiple criteria

        if op == "set":
            number_set = NeoSchema.update_data_node(data_node=uri, set_dict={"pinned": True})
        elif op == "unset":
            number_set = NeoSchema.update_data_node(data_node=uri, set_dict={"pinned": False})
        else:
            raise Exception("pin_category(): the argument `op` must be equal to either 'set' or 'unset'")

        assert number_set == 1, "pin_category(): no change could be made to the database"



    @classmethod
    def is_pinned(cls, uri :str) -> bool:
        """
        Return True if the given Category has a "pinned" status; otherwise, False

        :param uri: The URI of a data node representing a Category
        :return:    True or False
        """
        all_props = NeoSchema.search_data_node(uri=uri, labels="Categories")    # Returns a dict, or None
        assert all_props, "is_pinned(): unable to locate the specified Category node"

        value = all_props.get("pinned", False)  # Unless specifically "pinned", all Categories aren't

        return value



    @classmethod
    def get_see_also(cls, from_category :str) -> [dict]:
        """
        From the given Category, follow all the "see also" links, and return data about
        the Categories that are on their receiving end

        :param from_category:   URI of the Category where the "see_also" relationship originates
        :return:                List (possibly empty) of dictionaries of data from Categories linked to
                                    by a "see_also" relationship.  They keys are "name", "remarks", "uri".
                                    EXAMPLE: [{'name': 'Quotes', 'uri': '823', 'remarks': None}]
        """
        # TODO: perhaps restore the old feature of also storing a "description" field on the relationships

        return NeoSchema.follow_links(class_name="Categories", node_id=from_category, id_key="uri",
                                      link_name="BA_see_also", labels="Categories", properties=["name", "remarks", "uri"])



    @classmethod
    def create_see_also(cls, from_category :str, to_category :str) -> None:
        """
        Create a "see_also" link between the given Categories, in the specified direction

        :param from_category:   URI of the Category where the "see_also" relationship originates
        :param to_category:     URI of the Category where the "see_also" relationship terminates
        :return:                None
        """
        NeoSchema.add_data_relationship(from_id=from_category, to_id=to_category, id_type="uri",
                                        rel_name="BA_see_also")



    @classmethod
    def remove_see_also(cls, from_category :str, to_category :str) -> None:
        """
        Remove a "see_also" link between the given Categories, in the specified direction.
        If not found, an Exception is raised.

        :param from_category:   URI of the Category where the "see_also" relationship originates
        :param to_category:     URI of the Category where the "see_also" relationship terminates
        :return:                None
        """
        NeoSchema.remove_data_relationship(from_id=from_category, to_id=to_category, id_type="uri",
                                           rel_name="BA_see_also", labels="Categories")



    @classmethod
    def import_ontology(cls, df :pd.DataFrame, root_uri :str) -> int:
        """
        Import an ontology of Categories, from a Pandas dataframe whose columns, named "0", "1", "2", ...
        represent levels of increasing depth.
        Each row of the dataframe must have exactly 1 entry, representing a Category name;
        all other entries are expected to be NaN

        EXAMPLE of dataframe (NaN's not shown):
            0           1           2
            Chapter 1
                        Section 1
                                    Paragraph 1
                                    Paragraph 2
                                    Paragraph 3
                        Section 2
                                    Paragraph 1

        Note: the above dataframe may be imported from a header-free CSV file with operations such as
                    pd.read_csv('C:/my_folder/my_file.csv',
                                header=None, encoding = "ISO-8859-1")

        Note: all Category nodes get assigned a unique URI
              based on the namespace currently used by add_subcategory()

        :param df:          A Pandas dataframe with the data to import
        :param root_uri:    A string with the unique URI of an existing Category node that is to be
                                the parent of all the top-level (column "0") imported Categories
        :return:            The number of Category nodes created
        """
        n_cols = len(df.columns)            # Number of columns in the dataframe

        ancestry_name = [None] * n_cols     # A list with None repeated as many times as the number of columns
        ancestry_uri =  [None] * n_cols

        n_created = 0       # Number of new Category nodes created

        for ind in df.index:
            print("-- Row: ", ind, " --")
            s = df.loc[ind]      # a Pandas Series for the current row
            #print(s)

            # Find the first non-NaN index
            #TODO: skip blank entries alongside the NaN's
            level = int(s.first_valid_index())  # Note that the index will be either "0", "1", etc.

            # Get the value at the first non-NaN index
            category_name = s[level]

            print(f"Category: `{category_name}`, at level {level}")

            if level == 0:
                parent_name = "ROOT"
                parent_uri = root_uri
            else:
                parent_name = ancestry_name[level -1]
                parent_uri   = ancestry_uri[level -1]


            print(f"    CREATING `{category_name}` as a subcategory of `{parent_name}` (uri: '{parent_uri}')")
            #new_uri = "TEMP"
            new_uri = Categories.add_subcategory(category_uri=parent_uri,
                                                 data_dict={"subcategory_name": category_name} )
            n_created += 1
            print(f"        newly-assigned uri: '{new_uri}'")
            ancestry_name[level] = category_name
            ancestry_uri[level] = new_uri

            print()

        return n_created





    #####################################################################################################

    '''                                ~   VIEW ITEMS IN CATEGORIES   ~                                '''

    def ________VIEW_ITEMS_IN_CATEGORIES________(DIVIDER):
        pass        # Used to get a better structure view in IDEs
    #####################################################################################################

    @classmethod
    def get_categories_linked_to_content_item(cls, item_uri :str) -> [{}]:
        """
        Locate and return information about all the Categories
        that the given Content Item is linked to

        :param item_uri:    The URI of a data node representing a Content Item
        :return:            A list of dicts that have the keys "uri", "name", "remarks";
                                any missing value will appear as None
        """
        q = '''
            MATCH (:BA {uri: $item_uri}) - [:BA_in_category] -> (cat :Categories)
            RETURN cat.uri AS uri, cat.name AS name, cat.remarks AS remarks
            '''
        result = cls.db.query(q, data_binding={"item_uri": item_uri})
        return result



    @classmethod
    def get_content_items_by_category(cls, uri) -> [{}]:
        """
        Return the records for all nodes linked
        to the Category node identified by its uri value

        :param uri: A string identifying the desired Category
        :return:    A list of dictionaries
                    EXAMPLE:
                    [{'schema_code': 'i', 'uri': '1','width': 450, 'basename': 'my_pic', 'suffix': 'PNG', pos: 0, 'class_name': 'Images'},
                     {'schema_code': 'h', 'uri': '1', 'text': 'Overview', pos: 10, 'class_name': 'Headers'},
                     {'schema_code': 'n', 'uri': '1', 'basename': 'overview', 'suffix': 'htm', pos: 20, 'class_name': 'Notes'},
                     {'schema_code': 'rs', 'class_name': 'Recordset', 'class_handler': 'recordsets', 'uri': '6965', 'pos': 86, 'n_group': '4', 'order_by': 'name', 'class': 'YouTube Channel'}
                    ]
        """

        # Locate all the Content Items linked to the given Category, and also extract the name of the schema Class they belong to
        # TODO: switch to using one of the Collections methods

        q = '''
            MATCH (cl :CLASS)<-[:SCHEMA]- (n) -[r :BA_in_category]-> (:Categories {uri:$category_id})
            RETURN n, r.pos AS pos, cl.name AS class_name, cl.handler AS class_handler, cl.code AS class_code
            ORDER BY r.pos
            '''

        # TODO: class_code is being phased out in favor of the new class_handler

        result = cls.db.query(q, data_binding={"category_id": uri})
        #cls.db.debug_query_print(q, data_binding={"category_id": uri})


        content_item_list = []
        for elem in result:
            item_record = elem["n"]             # A dictionary with the various fields

            # TODO: eliminate possible conflict if the node happens to have
            #       attributes named "pos" or "class_name"!
            #       Ought to return elements that are dictionaries such as:
            #           {"pos": 15, "class_name": "Recordset", "class_handler": "recordsets", data: {}}

            item_record["pos"] = elem["pos"]                # Inject into the record a positional value
            item_record["class_name"] = elem["class_name"]  # Inject into the record the name of its Class

            if elem.get("class_handler"):
                item_record["class_handler"] = elem["class_handler"]    # Inject into the record the handler of its Class (not always present)

            if elem.get("class_code"):
                item_record["schema_code"] = elem["class_code"]         # Inject into the record the Class code (renamed "schema_code")
                                                                        # TODO: temp, during phaseout of "schema_code" in favor of "class_handler"
            #if ("schema_code" in elem) and (elem["schema_code"]) and ("schema_code" not in item_record):
                #item_record["schema_code"] = elem["schema_code"]


            if "date_created" in item_record:   # TODO: this is a hack, to clean up!
                del item_record["date_created"] # Datetime objects aren't serializable and lead to Flask errors
                                                # TODO: let NeoAccess handle the conversion to string
                                                # TODO: utilize a "type" attribute in the Schema Property node,
                                                #       to inform of the "datetime" data type

            content_item_list.append(item_record)

        #print(content_item_list)
        return content_item_list




    #####################################################################################################

    '''                        ~   ADD/REMOVE ITEMS FROM CATEGORIES   ~                               '''

    def ________ADD_REMOVE_ITEMS_FROM_CATEGORIES________(DIVIDER):
        pass        # Used to get a better structure view in IDEs
    #####################################################################################################

    @classmethod
    def add_content_at_beginning(cls, category_uri :str, item_class_name: str, item_properties: dict, new_uri=None, namespace="data_node") -> str:
        """
        Add a new Content Item, with the given properties and Class, to the beginning of the specified Category.

        TODO: solve the concurrency issue - of multiple requests arriving almost simultaneously, and being handled by a non-atomic update,
              which can lead to incorrect values of the "pos" relationship attributes.
              -> Follow the new way it is handled in add_content_at_end()

        :param category_uri:    The string "uri" of the Category to which this new Content Media is to be attached
        :param item_class_name: For example, "Images"
        :param item_properties: A dictionary with keys such as "width", "height", "caption","basename", "suffix" (TODO: verify against schema)
        :param new_uri:         Normally, if None (default) the Item ID is auto-generated,
                                    but it can also be provided (if provided, it MUST be unique)
        :param namespace:       Only applicable if new_uri is None : the namespace to use for automatically generating a URI
        :return:                The auto-increment "uri" assigned to the newly-created data node
        """
        if new_uri is None:
            # If a URI was not provided for the newly-created node,
            # then auto-generate it: obtain (and reserve) the next auto-increment value in the given namespace
            new_uri = NeoSchema.reserve_next_uri(namespace=namespace)    # Returns a string


        new_uri = Collections.add_to_collection_at_beginning(collection_uri=category_uri, membership_rel_name="BA_in_category",
                                                             item_class_name=item_class_name, item_properties=item_properties,
                                                             new_uri=new_uri)
        return new_uri



    @classmethod
    def link_content_at_end(cls, category_uri :str, item_uri :str) -> None:
        """
        Given an EXISTING data node, link it to the end of the specified Category.
        If a link to that Category already exists, an Exception is raised.

        :param category_uri:String to identify an existing Category
        :param item_uri:    String to identify an existing Content Item
        :return:            None
        """
        #TODO: verify that the item_uri is not referring to a Category!
        #      More generally, verify that its Class has a "BA_in_category" to
        #      the "Category" Class; this ought to be enforced by link_to_collection_at_end()

        # Link the Content Item to the end of the Category
        Collections.link_to_collection_at_end(item_uri=item_uri, collection_uri=category_uri,
                                              membership_link_name="BA_in_category")



    @classmethod
    def add_content_at_end(cls, category_uri :str, item_class_name: str, item_properties: dict, new_uri=None, namespace="data_node") -> str:
        """
        Add a NEW Content Item, with the given properties and Class, to the end of the specified Category collection.
        First, create a new Data Node, and then link it to the given Category, positioned at the end.

        :param category_uri:    A string to identify the Category
                                    to which this Content Media being newly-created is to be attached
        :param item_class_name: For example, "Images"
        :param item_properties: A dictionary with keys such as "width", "height", "caption","basename", "suffix"
                                    NOTE: if the Class was declared as "strict",
                                          then any key not declared in the Schema gets silently ignored
        :param new_uri:         Normally, if None (default) the Item ID is auto-generated,
                                    but it can also be provided (if provided, it MUST be unique)
        :param namespace:       Only applicable if new_uri is None : the namespace to use for automatically generating a URI
        :return:                The "uri" (passed or created) of the newly-created data node
        """
        if new_uri is None:
            # If a URI was not provided for the newly-created node,
            # then auto-generate it: obtain (and reserve) the next auto-increment value in the given namespace
            new_uri = NeoSchema.reserve_next_uri(namespace=namespace)    # Returns a string


        NeoSchema.create_data_node(class_node=item_class_name, properties=item_properties,
                                   extra_labels="BA", new_uri=new_uri,
                                   silently_drop=True)
        # NOTE: properties such as  "basename", "suffix" are stored with the Image or Document node,
        #       NOT with the Content Item node ;
        #       this is allowed by our convention about "INSTANCE_OF" relationships

        #print(f"add_content_at_end(): Created new Data Node with new_internal_id = {new_internal_id} and new_uri = '{new_uri}'")

        cls.link_content_at_end(category_uri=category_uri, item_uri=new_uri)

        return new_uri



    @classmethod
    def add_content_after_element(cls, category_uri :str, item_class_name: str, item_properties: dict, insert_after :str, new_uri=None, namespace="data_node") -> str:
        """
        Add a NEW Content Item, with the given properties and Class, inserted into the given Category after the specified Item
        (in the context of the positional order encoded in the relationship attribute "pos")

        :param category_uri:    String with a unique "uri" identified of the Category
                                    to which this new Content Media is to be attached
        :param item_class_name: For example, "Images"
        :param item_properties: A dictionary with keys specific to the new Content Item,
                                    such as "width", "height", "caption", "basename", "suffix" (TODO: verify against schema)
        :param insert_after:    The URI of the element after which we want to insert
        :param new_uri:         Normally, if None (default) the Item ID is auto-generated,
                                    but it can also be provided (if provided, it MUST be unique)
        :param namespace:       Only applicable if new_uri is None : the namespace to use for automatically generating a URI
        :return:                The auto-increment "uri" assigned to the newly-created data node
        """
        if new_uri is None:
            # If a URI was not provided for the newly-created node,
            # then auto-generate it: obtain (and reserve) the next auto-increment value in the given namespace
            new_uri = NeoSchema.reserve_next_uri(namespace=namespace)    # Returns a string


        # TODO: solve the concurrency issue - of multiple requests arriving almost simultaneously, and being handled by a non-atomic update,
        #       which can lead to incorrect values of the "pos" relationship attributes.
        #       -> Follow the new way it is handled in add_content_at_end()
        new_uri = Collections.add_to_collection_after_element(collection_uri=category_uri, membership_rel_name="BA_in_category",
                                                              item_class_name=item_class_name, item_properties=item_properties,
                                                              insert_after=insert_after,
                                                              new_uri=new_uri)
        return new_uri



    @classmethod
    def detach_from_category(cls, category_uri :str, item_uri :str) -> None:
        """
        Sever the link from the specified Content Item and the given Category.
        If it's the only Category that the Content Item is currently linked to,
        an Exception is raised (to avoid leaving that Content Item "stranded")

        :param category_uri:    The URI of a data node representing a Category
        :param item_uri:        The URI of a data node representing a Content Item
        :return:                None
        """
        match_from = cls.db.match(key_name="uri", key_value=item_uri)
        match_to = cls.db.match(labels="Categories")
        assert cls.db.number_of_links(match_from=match_from, match_to=match_to, rel_name="BA_in_category") > 1, \
            f"detach_from_category(): Cannot delete the only remaining 'BA_in_category' link " \
            f"from Content Item (URI: '{item_uri}') to Categories"

        NeoSchema.remove_data_relationship(from_id=item_uri, to_id=category_uri,
                                           rel_name="BA_in_category", labels=None)



    @classmethod
    def relocate_across_categories(cls, items :Union[List[str], str], from_category :str, to_category :str):
        """
        Given an existing list of data nodes (representing "Content Items" attached to the specified "from" Category),
        switch each of them to become a "Content Item" of the "to" Category, positioned at the end of it.

        The category-membership relationships ("BA_in_category") is severed from each the "Content Items" to the "from" Category,
        and a new one is created from that "Content Item" to the "to" Collection.

        Return the number of Content Items successfully relocated.

        :param items:           URI, or list of URI's, of Data Node(s)
                                    representing a "Content Items" attached to the "from" Category below
        :param from_category:   The URI of a Category Data Node to which the above Content Item(s) are connected
        :param to_category:     The URI of a Category Data Node to which the above Content Item(s) needs to be switched to
        :return:                The number of Content Items successfully relocated
        """
        return Collections.bulk_relocate_to_other_collection_at_end(items=items,
                                                             from_collection=from_category, to_collection=to_category,
                                                             membership_rel_name="BA_in_category")





    #####################################################################################################

    '''                                  ~   SCHEMA-RELATED    ~                                      '''

    def ________SCHEMA_RELATED________(DIVIDER):
        pass        # Used to get a better structure view in IDEs
    #####################################################################################################

    @classmethod
    def get_items_schema_data(cls, category_uri :str, exclude_system=True) -> dict:
        """
        Locate all the schema Classes used by Content Items attached to the given Category,
        and return a dictionary with the Properties (in the Schema order) of each,
        including Properties of their "ancestor" Classes (ancestral thru "INSTANCE_OF" relationships)

        Properties marked as "system" are optionally excluded (default).

        :param category_uri:A string with the "uri" value to identify the desired Category
        :param exclude_system: [OPTIONAL] If True, Property nodes with the attribute "system" set to True will be excluded;
                                    default is True
        :return:            A dictionary whose keys are Class names (of Content Items attached to the given Category),
                            and whose values are the Properties (in their Schema order) of those Classes.
                            Properties declared in "ancestor" Classes (thru "INSTANCE_OF" relationships)
                            are also included.
                            Properties regarded as "system" ones, such as "uri", are excluded.
                            EXAMPLE:
                                {'Headers': ['text'],
                                 'Site Link': ['url', 'name', 'date', 'comments', 'rating', 'read'],
                                 'Notes': ['title', 'public', 'date_created', 'basename', 'suffix']
                                 'German Vocabulary': ['Gender', 'German', 'English', 'notes'],
                                 'Quote': ['quote', 'attribution', 'notes'],
                                 'Recordset': ['class', 'order_by', 'clause', 'n_group']
                                 }
        """
        # Locate the names of the Classes of all the Content Items attached to the given Category
        q = '''
            MATCH   (CLASS {name: "Categories"}) <-[:SCHEMA]- (cat :Categories {uri: $category_uri}) 
                    <-[:BA_in_category]- (content_item) -[:SCHEMA]-> (cl:CLASS) 
            RETURN DISTINCT cl.name AS class_name
            '''
        #cls.db.debug_query_print(q, data_binding={"category_uri": category_uri})

        class_list = cls.db.query(q, data_binding={"category_uri": category_uri},
                                  single_column="class_name")
        # EXAMPLE: ["French Vocabulary", "Site Link"]


        # Now extract all the Property fields, in the schema-stored order, of the above Classes
        records_schema_data = {}
        for class_name in class_list:
            prop_list = NeoSchema.get_class_properties(class_node=class_name,
                                                       include_ancestors=True, sort_by_path_len="ASC",
                                                       exclude_system=exclude_system)
            records_schema_data[class_name] = prop_list

        return records_schema_data



    #####################################################################################################

    '''                          ~   POSITION WITHIN CATEGORIES    ~                                  '''

    def ________POSITIONING________(DIVIDER):
        pass        # Used to get a better structure view in IDEs
    #####################################################################################################

    @classmethod
    def check_for_duplicates(cls, category_name :str) -> [str]:
        """
        This is a diagnostic method to check for problems.

        Look for duplicates values in the "pos" attributes
        of the "BA_in_category" links ending in the specified Category node

        :param category_name:   To identifying the Category of interest
        :return:                In case of duplicates, return a list of texts with error reports;
                                    if no duplicates, return an empty list
        """
        q = '''
            MATCH (ci1)-[r1:BA_in_category] -> (:Categories {name: $name}) <- [r2:BA_in_category]-(ci2) 
            WHERE r1.pos = r2.pos AND id(ci1) < id(ci2)
            RETURN r1.pos AS pos, ci1.uri AS item1, ci2.uri AS item2
            ORDER BY r1.pos
            '''

        data_binding = {"name": category_name}

        #cls.db.debug_query_print(q, data_binding)
        duplicates = cls.db.query(q, data_binding)

        if not duplicates:
            return []           # No problems found
        else:
            problem_list = []
            for record in duplicates:
                problem = f"Problem found: Duplicate 'pos' value ({record['pos']}), " \
                          f"shared by links to Content Items with URIs `{record['item1']}` and `{record['item2']}`"
                problem_list.append(problem)

            return problem_list



    @classmethod
    def check_all_categories_for_duplicates(cls) -> [dict]:
        """
        This is a diagnostic method to check for problems.

        :return:    A (possibly-empty) list of dicts, detailing all located duplicates
        """
        q = '''
            MATCH (ci1)-[r1 :BA_in_category]->(c :Categories)<-[r2 :BA_in_category]-(ci2)
            WHERE r1.pos = r2.pos AND id(ci1) < id(ci2)
            RETURN c.name AS category_name, r1.pos AS pos, ci1.uri AS uri1, ci2.uri AS uri2
            ORDER BY category_name, pos
            '''
        return cls.db.query(q)



    @classmethod
    def reassign_positional_values(cls, category_name :str) -> None:
        """
        To re-assign positional values of Content Items within the given Category
        (starting at 0 and proceeding in increments of Collections.DELTA_POS):

        :param category_name:
        :return:                None
        """
        q = f'''       
            MATCH (ci)-[r :BA_in_category]->(n :Categories {{name: $category_name}})           
            WITH r.pos AS POS, id(ci) AS NODE_ID           
            ORDER by r.pos
            
            WITH collect(NODE_ID) AS ID_LIST           
            WITH size(ID_LIST) AS TOT, ID_LIST           
            WITH range(0, TOT-1) AS INDEX_LIST, ID_LIST
            
            UNWIND INDEX_LIST AS i
            
            MATCH (x)-[newr :BA_in_category]->(n :Categories {{name: $category_name}}) WHERE id(x) = ID_LIST[i]
            
            SET newr.pos = i * {Collections.DELTA_POS}
            
            RETURN x.uri, i, newr.pos
        '''
        cls.db.query(q, data_binding={"category_name": category_name})



    @classmethod
    def reposition_content(cls, category_uri :str, uri: str, move_after_n: int):
        """
        Reposition the given Content Item after the n-th item (counting starts with 1) in specified Category.

        Note: there's no harm (though it's wasteful) to move an item to a final sequence position where it already is;
              its "pos" value will change

        :param category_uri:    A string identifying the desired Category
        :param uri:             A string with the URI of the Content Item we're repositioning
        :param move_after_n:    The index (counting from 1) of the item after which we want to position the item being moved
                                    Use n=0 to indicate "move before anything else"
        :return:
        """
        assert NeoSchema.is_valid_uri(category_uri), "ERROR: argument 'category_uri' is invalid"
        assert NeoSchema.is_valid_uri(uri), "ERROR: argument 'uri' is not a valid string"
        assert type(move_after_n) == int, "ERROR: argument 'move_after_n' is not an integer"
        assert move_after_n >= 0, "ERROR: argument 'move_after_n' cannot be negative"

        # Collect a subset of the first sorted "pos" values: enough values to cover across the insertion point
        number_to_consider = move_after_n + 1
        q = f'''
            MATCH (c:BA:Categories {{uri: $category_id}}) <- [r:BA_in_category] - (:BA)
            WITH  r.pos AS pos
            ORDER by pos
            LIMIT {number_to_consider}
            WITH collect(pos) AS POS_LIST
            RETURN POS_LIST
            '''

        result = cls.db.query(q, {"category_id": category_uri})  # If nothing found, this will be [{'POS_LIST': []}]
        pos_list = result[0].get("POS_LIST")    # A subset of the first sorted "pos" values
        #print("pos_list: ", pos_list)

        if pos_list == []:
            # The Category is empty (or doesn't exist)
            raise Exception(f"Category (id {category_uri}) not found, or empty")

        if move_after_n == 0:
            # Move to top
            #print("Moving to the top")
            top_pos = pos_list[0]
            new_pos = top_pos - cls.DELTA_POS
        elif move_after_n >= len(pos_list):
            # Move to bottom
            #print("Moving to the bottom")
            top_pos = pos_list[-1]      # The last element in the list
            new_pos = top_pos + cls.DELTA_POS
        else:
            pos_above = pos_list[move_after_n - 1]  # The "pos" value of the Item just above the insertion point
            pos_below = pos_list[move_after_n]      # The "pos" value of the Item just below
            #print(f"pos_above: {pos_above} | pos_below: {pos_below}")
            if pos_below == pos_above + 1:
                # There's no room; shift everything that is past that position, by a count of DELTA_POS
                #print(f"********* RELOCATING ITEMS (skipping the first {move_after_n}) ***********")
                cls.relocate_positions(category_uri, n_to_skip=move_after_n, pos_shift=cls.DELTA_POS)
                new_pos = pos_above + int(cls.DELTA_POS/2)			# This will be now be the empty halfway point
            else:
                new_pos = int((pos_above + pos_below) / 2)		# Take the halfway point, rounded down


        # Change the "pos" attribute of the relationship to the Content Item being moved
        q = f'''
            MATCH (:BA:Categories {{uri: $category_id}}) <- [r:BA_in_category] - (:BA {{uri: $uri}})
            SET r.pos = {new_pos}
            '''

        #print("q: ", q)

        result = cls.db.update_query(q, {"category_id": category_uri, "uri": uri})
        number_props_set = result.get('properties_set')
        #print("number_props_set: ", number_props_set)
        if number_props_set != 1:
            raise Exception(f"Content Item (id {uri}) not found in Category (id {category_uri}), or could not be moved")



    @classmethod
    def relocate_positions(cls, category_uri :str, n_to_skip: int, pos_shift: int) -> int:
        """
        Shift the values of the "pos" attributes on the "BA_in_category" relationships
        from the given Category node, by the given amount;
        however, SKIP the first n_to_skip entries (as sorted by the "pos" attribute)

        EXAMPLE - given the following order of the relationships attached to the given Category:
            pos
        1:   45
        2:   84
        3:   91

        then relocate_positions(category_uri, n_to_skip=1, pos_shift=100) will result in:
            pos
        1:   45     <= got skipped
        2:  184
        3:  191

        :param category_uri:A string identifying the desired Category
        :param n_to_skip:   The number of relationships (after sorting them by "pos") NOT to re-position
                                Must be an integer >= 1 (it'd be pointless to shift everything!)
        :param pos_shift:   The increment by which to shift the values of the "pos" attributes on the relationships

        :return:            The number of repositionings performed
        """
        assert NeoSchema.is_valid_uri(category_uri), "ERROR: argument 'category_uri' is not a valid string"
        assert type(n_to_skip) == int, "ERROR: argument 'n_to_skip' is not an integer"
        assert type(pos_shift) == int, "ERROR: argument 'pos_shift' is not an integer"
        assert n_to_skip >= 1, "ERROR: argument 'n_to_skip' must be at least 1"

        q = f'''
            MATCH (c:BA:Categories {{uri: $category_id}}) <- [r:BA_in_category] - (:BA)
            WITH  r.pos AS pos, r
            ORDER by pos
            SKIP {n_to_skip}
            SET r.pos = r.pos + {pos_shift}
            '''

        result = cls.db.update_query(q, {"category_id": category_uri})
        return result.get('properties_set')



    @classmethod
    def swap_content_items(cls, uri_1 :str, uri_2 :str, cat_id :str) -> None:
        """
        Swap the positions of the specified Content Items within the given Category

        :param uri_1:   A string with the uri of the 1st Content Item
        :param uri_2:   A string with the uri of the 2nd Content Item
        :param cat_id:  A string with the uri of the Category
        :return:        None
        """

        # Validate the arguments
        assert uri_1 != uri_2, \
            f"Attempt to swap a Content Item (URI `{uri_1}`) with itself!"


        # Look for a Category node (c) that is connected to 2 Content Items (n1 and n2)
        #   thru "BA_in_category" relationships; then swap the "pos" attributes on those relationships
        q = '''
            MATCH (n1 {uri: $uri_1})
                        -[r1:BA_in_category]->(c:BA:Categories {uri: $cat_id})<-[r2:BA_in_category]-
                  (n2 {uri: $uri_2})
            WITH r1.pos AS tmp, r1, r2
            SET r1.pos = r2.pos, r2.pos = tmp
            '''

        data_binding = {"uri_1": uri_1, "uri_2": uri_2, "cat_id": cat_id}

        #print(q)
        #print(data_binding)

        stats = cls.db.update_query(q, data_binding)
        #print("stats of query: ", stats)

        assert stats != {},  \
            f"Irregularity detected in swap action: Unable to determine the success of the operation"

        number_properties_set = stats.get("properties_set")

        assert number_properties_set, \
            f"Failure to swap content items `{uri_1}` and `{uri_2}` within Category `{cat_id}`. Query: {q}"

        assert number_properties_set == 2, \
            f"Irregularity detected in swap action: {number_properties_set} properties were set," \
            f" instead of the expected 2"




    #####################################################################################################

    '''                                  ~   PAGE HANDLER   ~                                         '''

    def ________PAGE_HANDLER________(DIVIDER):
        pass        # Used to get a better structure view in IDEs
    #####################################################################################################

    @classmethod
    def viewer_handler(cls, category_uri :str):
        """
        Handler function for the Flask page generator "BA_pages_routing.py"

        :param category_uri: A string identifying the desired Category
        :return:             A list of dictionaries, with one element for each "sibling";
                                each element contains the 'internal_id' and 'neo4j_labels' keys,
                                plus whatever attributes are stored on that node.
                                EXAMPLE of single element:
                                {'name': 'French', 'internal_id': 123, 'neo4j_labels': ['Categories', 'BA']}
        """
        # TODO: expand to cover all the data needs of BA_pages_routing.py
        # TODO: maybe move to DataManager layer

        category_internal_id = NeoSchema.get_data_node_internal_id(uri = category_uri)
        siblings_categories = Categories.get_sibling_categories(category_internal_id)

        return siblings_categories



import os
import re               # For REGEX
import unicodedata



class UploadHelper:
    """
    Helper class to manage file uploads with Flask 1.1
    """


    @classmethod
    def store_uploaded_file(cls, files, upload_dir: str, key_name=None) -> (str, str, str, str):
        """
        Meant for SINGLE-file uploads.

        It accepts a flask.request object, and it creates a file from the upload,
        which it stores into the folder specified by upload_dir

        It decides, from the name of the file being uploaded, a meaningful (and safe) name
        for the temporary file to create as part of the upload.

        It returns both the basename and the full filename of the temporary file thus created,
        as well as the original name of the file.

        In case of error, an Exception is raised

        TODO: maybe allow to optionally provide the final location/name for the uploaded file

        :param files:       An ImmutableMultiDict object.
                                EXAMPLE: ImmutableMultiDict([('imported_datafile', <FileStorage: 'my_data.json' ('application/json')>)])
                                    where 'imported_datafile' originates from <input type="file" name="imported_datafile">
                                    and the name after FileStorage is the name of the file being uploaded
        :param upload_dir:  Name of temporary directory to use for the upload.
                                EXAMPLES: "/tmp/" (Linux)  or  "D:/tmp/" (Windows)
        :param key_name:    (OPTIONAL) The value that was used in the HTML form INPUT tag for the "name" attribute:
                                <input type="file" name="some_key_to_refer_to_the_upload">
                                or its counterpart in JS submissions, such as
                                            const post_data = new FormData();
                                            post_data.append('file', this.file_to_import);
                            Basically, a label to tag the file being uploaded.
                            The "Dropzone" front-end module uses "file".
                            If not provided, the first key found in request.files is used.
                            (Note: in the ImmutableMultiDict data structure, multiple values are allowed for the same key;
                                   if more than one value is present, the first is picked)

        :return:            The 4-tuple (filename, full_filename_including_path, original_filename, mimetype)
                            where filename is the basename;
                            filename and original_filename are usually the same, unless there were
                            funky characters in original_filename
                            EXAMPLE: ("my_file_being_uploaded.txt", "D:/tmp/my_file_being_uploaded.txt",
                                      "my_file_being_uploaded.txt", "text/plain")

        """
        # Locate the data about the desired file upload
        if len(files) == 0:
            raise Exception(f"No files found in upload!  (The ImmutableMultiDict object in request.files is empty)")


        if key_name is not None:    # If a specific "name" (tag attached to uploaded file) was requested
            # Example from an invoking form:   <input type="file" name="some_key_to_refer_to_the_upload">
            if key_name not in files:
                raise Exception(f"No file tagged with the label `{key_name}` found in the upload!  "
                                f"Check the `name` attribute in the HTML `input` tag (it should be '{key_name}').")

        else:                       # If no particular "name" (tag) was specified, just pick the first one
            key_name = list(files)[0]
            print(f"store_uploaded_file(): No specific key (tag associated with uploaded file) requested. "
                  f"The first key found in request.files (`{key_name}`) will be used.")


        f = files[key_name]      # f is a Werkzeug FileStorage object
        # Note: if multiple uploads were attached to this key, only the first one is picked

        #print("f: ", f)                    # EXAMPLE: <FileStorage: 'abc.txt' ('text/plain')>
        #print("f.stream: ", f.stream)      # EXAMPLE: <tempfile.SpooledTemporaryFile object at 0x000001604F5E6700>
        #print("f.filename: ", f.filename)  # EXAMPLE: abc.txt
        #print("f.mimetype: ", f.mimetype)  # EXAMPLE: text/plain

        # If the user did not select a file, the browser submits an
        #       empty file without a filename
        if f.filename == '':
            raise Exception("No selected file!  Did you select a file to upload?")


        # Construct, from the name of the file being uploaded, a "safe" name (free of funky characters)
        #   for the temporary file to create as part of the upload
        tmp_filename_for_upload = cls.secure_filename_BA(f.filename)
        #print(f"Name given to temporary file to upload to: `{tmp_filename_for_upload}`")

        #print(f"Attempting to upload to directory `{upload_dir}`")

        full_filename = f"{upload_dir}{tmp_filename_for_upload}"    # EXAMPLE: "D:/tmp/my_file_being_uploaded.txt"
        f.save(full_filename)                                       # Create the temporary file, which concludes the upload

        return (tmp_filename_for_upload, full_filename, f.filename, f.mimetype)    # Normal termination



    @classmethod
    def secure_filename_BA(cls, filename: str) -> str:
        """
        Return a secure version of a filename.
        This filename can then safely be stored on a regular file system and passed
        to os.path.join.  The filename returned is an ASCII only string
        for maximum portability.

        ADAPTED FOR BRAIN ANNEX FROM werkzeug.utils.secure_filename(), version 0.5;
        blank spaces, commas, ampersand, and round parentheses are now treated as valid.
        However, blank spaces at start/end of names are still dropped.
        See: https://flask.palletsprojects.com/en/2.0.x/patterns/fileuploads/

        On Windows, the function also makes sure that the file is not
        named after one of the special device files.

        EXAMPLES:   secure_filename_BA("My cool ++ movie.mov")          -> 'My cool  movie.mov'
                    secure_filename_BA("../../../etc/passwd")           -> 'etc_passwd'
                    secure_filename_BA('i contain \xfcml\xe4uts.txt')   -> 'i contain umlauts.txt'
                    secure_filename_BA("    blank-padded string  ")     -> 'blank-padded string'
                    secure_filename_BA("COM1.txt")       [On Windows]   -> '_COM1.txt'

        WARNING: The function might return an empty filename.  It's your responsibility
        to ensure that the filename is unique and that you abort or
        generate a random filename if the function returned an empty one.

        :param filename:    A string with the "cleaned-up" filename to use
        """
        #TODO: pytest
        _filename_ascii_strip_re = re.compile(r"[^A-Za-z0-9_., &()-]") # List of allowed characters in the name;
                                                                       # BRAIN ANNEX adaptation:
                                                                       # note that the blank space, ampersand and the round parentheses
                                                                       # are included
        _windows_device_files = (
            "CON",
            "AUX",
            "COM1", "COM2", "COM3", "COM4",
            "LPT1", "LPT2", "LPT3",
            "PRN",
            "NUL"
        )

        # Convert non-ASCII characters to closest ASCII equivalent.  EXAMPLE: "Rüdiger" -> "Rudiger"
        filename = unicodedata.normalize("NFKD", filename)  # Normal form "KD" (Compatibility Decomposition)
        filename = filename.encode("ascii", "ignore").decode("ascii")

        # Replace OS file path separators (such as forward and back slashes) with underscores.  EXAMPLE: "bin/src" -> "bin_src"
        for sep in os.path.sep, os.path.altsep:
            if sep:
                filename = filename.replace(sep, "_")

        #filename = "_".join(filename.split()   # Replace all blanks spaces with underscores
        filename = _filename_ascii_strip_re.sub("", filename)   # Zap all characters except the allowed ones
        filename = filename.strip("._ ")         # Zap periods, underscores and blanks at the start or end of the string

        # On nt (such as Windows 7, etc) a couple of special files are present in each folder.  We
        # have to ensure that the target file is not such a filename; if it is, we prepend an underline
        if  (
                os.name == "nt"
                and filename
                and filename.split(".")[0].upper() in _windows_device_files
        ):
            filename = f"_{filename}"

        return filename



__version__ = "5.0.0-beta.47"


from brainannex.neo_schema.neo_schema import NeoSchema
from brainannex.data_manager import DataManager
from brainannex.user_manager import UserManager
from brainannex.categories import Categories
from brainannex.collections import Collections
from brainannex.full_text_indexing import FullTextIndexing
import brainannex.PLUGINS.plugin_support as plugin_support


__all__ = [
    'NeoSchema',
    'DataManager',
    'UserManager',
    'Categories',
    'Collections',
    'FullTextIndexing',
    'plugin_support',
    'version'
]


def version():
    return __version__



# Utilities for comparisons

import collections


def compare_recordsets(rs1: [{}], rs2: [{}]) -> bool:
    """
    We define "recordsets" as "lists of dictionaries".  Each element of the lists is regarded as a "record".

    EXAMPLE of recordset:  [{'Field_A': 1},
                            {'Field_A': 1},
                            {'Field_A': 99, 'Field_B': 'hello'}]

    Compare 2 recordsets WITHOUT REGARD to the position of the dictionaries within the lists,
    and also WITHOUT REGARD to the position of the key:value pairs within the dictionaries.
    Duplicates records, if present, are treated as completely separate.

    Return True if the given recordsets match, as defined above; False, otherwise.

    WARNING: this function is meant for comparing SMALL datasets, because it's Order n square!

    :param rs1: A (possibly empty) list of dictionaries
    :param rs2: A (possibly empty) list of dictionaries

    :return:    True if there's a match, or False otherwise
    """

    # Verify the type of the arguments
    assert isinstance(rs1, list), "compare_recordsets() : The 1st argument is not a list!  Value = " + str(rs1)
    assert isinstance(rs2, list), "compare_recordsets() : The 2nd argument is not a list!  Value = " + str(rs2)

    if len(rs1) != len(rs2):
        return False    # Datasets of different sizes will never match

    rs2_clone = rs2.copy()  # IMPORTANT: this is done to avoid altering the list that was passed as argument

    # Consider each element (i.e. a dictionary) in turn in the first list:
    #   attempt to remove it from the other list; if the removal fails, then it means
    #   that we have an element in the 1st list that is not present in the 2nd one (hence a mismatch)
    for rec1 in rs1:
        # Note: since Python 3.7 dictionaries are order-preserving, but
        #       built-in Python functions such as "remove"
        #       do not distinguish dictionaries based on order:
        #           {'a': 1, 'b': 2} will match {'b': 2, 'a': 1}

        try:
            rs2_clone.remove(rec1)    # Remove (the first instance of) the element rec1 from the list rs2
        except Exception:
            return False        # The remove failed - i.e. the first list contains an element not in the 2nd one

    return True



def compare_unordered_lists(l1: [], l2: []) -> bool:
    """
    Compare two lists regardless of order of elements.
    Duplicates elements, if present, are treated as completely separate.
    IMPORTANT:  the elements of the list must match exactly,
                and *must* be HASHABLE Python entities, such as
                strings, numbers or tuples; they can NOT be, for example, dictionaries

    Return True if the given lists match, as defined above; False, otherwise.

    EXAMPLES:   [1, 2, 3] will match [3, 2, 1]
                [] and [] will match
                ["x", (1, 2)] will match [(1, 2) , "x"] but NOT ["x", (2, 1)]
                ["a", "a"] will NOT match ["a"]

    :param l1:  A list of HASHABLE Python entities (e.g. strings or numbers)
    :param l2:  Same as above
    :return:    True if there's a match, or False otherwise
    """
    return collections.Counter(l1) == collections.Counter(l2)



import re
import html
from typing import Union, List, Set
from neoaccess.cypher_utils import CypherUtils
from brainannex.neo_schema.neo_schema import NeoSchema
import brainannex.utilities.exceptions as exceptions


class FullTextIndexing:
    """
    Indexing-related methods, for full-text searching.

    For more info and background info, please see:
        https://julianspolymathexplorations.blogspot.com/2023/08/full-text-search-neo4j-indexing.html

    NOTE: no stemming nor lemmatizing is done.
          Therefore, for best results, all word searches should be done on stems;
          for example, search for "learn" rather than "learning" or "learns" - to catch all 3
    """

    # The "db" class properties gets set by InitializeBrainAnnex.set_dbase()
    db = None           # Object of class "NeoAccess".  MUST be set before using this class!
                        # TODO: add a method set_database(), as done for NeoSchema


    TAG_RE = re.compile(r'<[^>]+>')         # Use regex to strip off all HTML

    # List of common English words to skip from indexing (stopword list)
    # See https://github.com/Alir3z4/stop-words
    # TODO: allow over-ride in config file
    # Note: 2-letter words are redundant from the list below, because those don't get indexed
    COMMON_WORDS = ['and', 'or', 'either', 'nor', 'neither',
                    'the', 'an', 'with', 'without', 'within',
                    'in', 'out', 'on', 'off', 'at', 'of', 'from', 'to', 'into', 'not', 'but', 'by', 'upon',
                    'if', 'whether', 'then', 'else',
                    'me', 'my', 'mine', 'he', 'she', 'it', 'him', 'his', 'her', 'its',
                    'we', 'us', 'our', 'you', 'your', 'yours', 'they', 'them', 'their',
                    'why', 'because', 'due', 'since', 'until', 'through', 'thru',
                    'how', 'for', 'both', 'indeed',
                    'help', 'helps', 'let', 'lets',
                    'go', 'goes', 'going', 'gone', 'went',
                    'became', 'become', 'becomes', 'come', 'comes', 'came',
                    'be', 'is', 'isn', 'am', 'are', 'aren', 'were', 'been', 'was', 'wasn', 'being',
                    'can', 'cannot', 'could', 'might', 'may', 'do', 'does', 'doesn', 'did', 'didn', 'done', 'doing',
                    'make', 'made', 'making',
                    'have', 'haven', 'has', 'had', 'hadn', 'having',
                    'must', 'need', 'needs', 'needed', 'seem', 'seems', 'seemed', 'want', 'wants', 'should', 'shouldn', 'ought',
                    'will', 'would', 'shall',
                    'get', 'gets', 'got', 'give', 'gives', 'gave', 'giving',
                    'take', 'takes', 'took', 'taking', 'put', 'bring', 'brings', 'bringing',
                    'see', 'sees', 'given', 'end', 'start', 'starts', 'starting',
                    'ask', 'asks', 'answer', 'answers', 'question', 'questions',
                    'when', 'where', 'which', 'who', 'why', 'what',
                    'no', 'non', 'not', 'yes', 'maybe', 'perhaps', 'ok', 'oh',
                    'ie', 'i.e', 'eg', 'e.g',
                    'll', 've', 'so',
                    'good', 'better', 'best', 'well', 'bad',  'worse', 'worst',
                    'just', 'about', 'above', 'again', 'ago',
                    'times', 'date', 'dates', 'today', 'day', 'month', 'year', 'yr', 'days', 'months', 'years',
                    'hour', 'hr', 'minute', 'min', 'second', 'sec', 'pm',
                    'now', 'currently', 'late', 'early', 'soon', 'later', 'earlier', 'already',
                    'after', 'before', 'prior', 'yet', 'whenever', 'while', 'during', 'ever',
                    'follow', 'follows', 'following', 'along',
                    'never', 'seldom', 'rarely', 'occasionally', 'sometimes',
                    'often', 'always', 'usually', 'eventually', 'typical', 'typically', 'everyday',
                    'almost', 'quite', 'nearly',
                    'frequent', 'ubiquitous', 'usual', 'common', 'commonly',
                    'remarkable', 'impressive',
                    'really', 'approximately',
                    'allow', 'allows', 'allowing',
                    'old', 'older', 'new', 'newer', 'recent', 'recently',
                    'begin', 'began', 'start', 'starting', 'started',
                    'in', 'out', 'here', 'there',
                    'instead', 'alternative', 'alternatively', 'otherwise', 'case', 'cases', 'extent',
                    'up', 'down', 'over', 'above', 'under', 'below', 'between', 'among', 'wherever',
                    'next', 'previous', 'other', 'others', 'another', 'thing', 'things',
                    'like', 'as', 'aka', 'akin', 'such', 'fairly', 'actual', 'actually',
                    'likewise', 'similar', 'similarly',
                    'simple', 'simpler', 'simplest', 'simply',
                    'each', 'per', 'any', 'all', 'every', 'everyone', 'anyone', 'anybody', 'anything', 'something', 'someone', 'some', 'various',
                    'more', 'most', 'mostly', 'additional', 'extra',
                    'less', 'least', 'than', 'enough', 'only', 'further',
                    'everything', 'nothing',
                    'few', 'fewer', 'many', 'multiple', 'several', 'much', 'same', 'equal', 'different', 'unlike',
                    'full', 'empty', 'lot', 'very', 'around', 'vary', 'varying',
                    'complete', 'incomplete', 'thorough', 'thoroughly',
                    'approximately', 'approx', 'somewhat', 'certain', 'uncertain',
                    'directly', 'indirectly',
                    'part', 'parts', 'wide', 'broad', 'narrow', 'side', 'across',
                    'hence', 'therefore', 'thus', 'whereas', 'nevertheless', 'notwithstanding',
                    'whom', 'whoever', 'whomever', 'whatever', 'whose',
                    'this', 'that', 'these', 'those',
                    'too', 'also',
                    'related', 'issues', 'issue', 'referred',
                    'type', 'types', 'instance', 'instances',
                    'use', 'uses', 'used', 'using',
                    'com', 'org', 'www', 'http', 'https',
                    'zero', 'one', 'ones', 'two', 'three', 'four', 'five', 'six', 'seven', 'eight', 'nine',
                    'first', 'second', 'third', 'fourth', 'last',
                    'iii',
                    'include', 'including', 'incl', 'except', 'sure', 'according', 'accordingly',
                    'example', 'examples', 'define', 'defined',
                    'basically', 'essentially', 'finally',
                    'called', 'named', 'name', 'consider', 'considering', 'however', 'especially', 'etc',
                    'happen', 'happens', 'happening', 'continue', 'continues', 'continuing',
                    'change', 'changes', 'changing', 'changed',
                    'small', 'smaller', 'smallest', 'little', 'brief', 'briefly',
                    'short', 'shorter', 'shortest',
                    'big', 'bigger', 'biggest', 'large', 'larger', 'largest', 'huge',
                    'long', 'longer', 'longest',
                    'great', 'greater', 'greatest',
                    'low', 'lower', 'lowest', 'high', 'higher', 'highest', 'limited',
                    'increase', 'increased', 'decrease', 'decreased', 'vary', 'varies', 'varying',
                    'consist', 'consists', 'consisting', 'result', 'results', 'resulting',
                    'description', 'descriptions', 'describe', 'describing',
                    'hello', 'hi', 'info',
                    're', 'vs', 'ex',
                    'data', 'value', 'values',
                    'obvious', 'obviously', 'clearly',
                    'show', 'shows', 'showing', 'find', 'finds', 'found', 'finding', 'findings', 'respectively',
                    'still', 'size', 'pre', 'inc', 'comfortably', 'look', 'approach',
                    'exact', 'exactly', 'likely', 'probable', 'probably', 'avg', 'total', 'misc',
                    'require', 'requires', 'requiring', 'quick', 'quickly', 'rather',
                    'know', 'knows', 'knowing', 'say', 'says', 'saying']


    # TODO: allow user-specific words, from a configuration file.  For example, for German: ich, du, er, sie, wir, ihr



    ##########   STRING METHODS   ##########

    @classmethod
    def split_into_words(cls, text: str, to_lower_case=True, drop_html=True) -> [str]:
        """
        Lower-level function used in the larger context of indexing text that may contain HTML.

        Given a string, optionally zap HTML tags and HTML entities, such as &ndash;
        then ditch punctuation from the given text;
        finally, break it up into individual words, returned as a list.  If requested, turn it all to lower case.

        Care is taken to make sure that the stripping of special characters does NOT fuse words together;
        e.g. avoid turning 'Long&ndash;Term' into a single word as 'LongTerm';
        likewise, avoid turning 'One<br>Two' into a single word as 'OneTwo'

        EXAMPLE.  Given:
            '<p>Mr. Joe&amp;sons<br>A Long&ndash;Term business! Find it at &gt; (http://example.com/home)<br>Visit Joe&#39;s &quot;NOW!&quot;</p>'
            It will return (if to_lower_case is False):
            ['Mr', 'Joe', 'sons', 'A', 'Long', 'Term', 'business', 'Find', 'it', 'at', 'http', 'example', 'com', 'home', 'Visit', 'Joe', 's', 'NOW']

        Note about numbers:  * negative signs are lost  * numbers with decimals will get split into two parts

        :param text:            A string with the text to parse
        :param to_lower_case:   If True, all text is converted to lower case
        :param drop_html:       Use True if passing HTML text
        :return:                A (possibly empty) list of words in the text,
                                    free of punctuation, HTML and HTML entities such as &ndash;
        """
        # TODO: maybe eliminate the decimal numbers while leaving the integers alone, to allow indexing of (some) integer numbers such as years

        if drop_html:
            unescaped_text = html.unescape(text)    # Turn HTML entities into characters; e.g. "&ndash;" into "-"
            #print(unescaped_text)  # <p>Mr. Joe&sons<br>A Long–Term business! Find it at > (http://example.com/home)<br>Visit Joe's "NOW!"</p>

            stripped_text = cls.TAG_RE.sub(' ', unescaped_text) # Use regex to strip off all HTML and turn each occurrence into a blank
            #print(stripped_text)   # Mr. Joe&sons A Long–Term business! Find it at > (http://example.com/home) Visit Joe's "NOW!"    <- Note: blank space at end
        else:
            stripped_text = text


        if to_lower_case:
            stripped_text = stripped_text.lower()   # If requested, turn everything to lower case

        result = re.findall(r'\w+', stripped_text)  # Use regex to split off the string into individual words
        # (note that the "&" in the original string is treated as blank space/word separator)
        # EXAMPLE:  ['Mr', 'Joe', 'sons', 'A', 'Long', 'Term', 'business', 'Find', 'it', 'at', 'http', 'example', 'com', 'home', 'Visit', 'Joe', 's', 'NOW']

        return result



    @classmethod
    def extract_unique_good_words(cls, text :str, drop_html=True) -> Set[str]:
        """
        Higher-level function to prepare text for indexing;
        use the drop_html flag if the text contains HTML.

        From the given text, it returns the set of "acceptable", unique words.
        It does the following:

            1) zap punctuation
            2) if requested, HTML, HTML entities (such as &ndash;);
            3) turn into lower case
            4) break up into individual words
            5) strip off leading/trailing underscores
            6) eliminate "words" that match at least one of these EXCLUSION test:
                * are just 1 or 2 characters long
                * are numbers
                * start with digits (e.g. "50m" or "123plus")
                * are found in a list of common words

            7) eliminate duplicates

        Note: no stemming or other grammatical analysis is done.

        EXAMPLE - given
                  '<p>Mr. Joe&amp;sons<br>A Long&ndash;Term business! Find it at &gt; (http://example.com/home)<br>Visit Joe&#39;s &quot;NOW!&quot;</p>'
                  it returns:
                  ['mr', 'joe', 'sons', 'long', 'term', 'business', 'find', 'example', 'home', 'visit']

        :param text:        A string with the text to analyze and index
        :param drop_html:   Use True if passing HTML text
        :return:            A (possibly empty) set of strings containing "acceptable",
                                unique words in the text
        """

        assert type(text) == str, \
            f"extract_unique_good_words(): the argument must be a string; instead, it was of type {type(text)}"

        # Extract words from string, and convert them to lower case
        split_text = cls.split_into_words(text, to_lower_case=True, drop_html=drop_html)
        #print("The split text is : ", split_text)

        # Eliminate "words" that are 1 or 2 characters long, or that are numbers,
        # or that are in a list of common words.  Also eliminate duplicates
        word_set = set()    # Empty set

        # Define a regular expression pattern to match anything that STARTS with a numeric character
        pattern = r"^\d"        # Formerly: pattern = r"\d"

        for word in split_text:
            word = word.strip("_")
            if len(word) > 2 \
                    and not word.isnumeric() \
                    and word not in cls.COMMON_WORDS \
                    and not re.findall(pattern, word):
                word_set.add(word)      # Add the element to the set, if it passed all the exclusions

        #print("The word set for the index is: ", word_set)

        return word_set




    #############   GRAPH METHODS   #############

    @classmethod
    def initialize_schema(cls, content_item_class_name="Content Item") -> None:
        """
        Initialize the graph-database Schema used by this Indexer module:

        1) It will create a new "Word" Class linked to a new "Indexer" Class,
        by means of an outbound "occurs" relationship.
        The newly-created "Word" Class will be given one Property: "name".

        2) It will add a relationship named "has_index" from an existing (or newly created)
        "Content Item" Class to the new "Indexer" Class.

        NOTE: if an existing Class with the named specified by the argument `content_item_class_name`
              is not found, it will be created with some default values

        :param content_item_class_name: (OPTIONAL) The name of the Schema Class for Content Items,
                                            i.e. the Class of the Data Items to be indexed
                                            if not found, it gets created
                                            EXAMPLES: "Document", "Notes", "Content Items" (default)
        :return:                        None
        """
        #         TODO: manage indexing.  If done in Cypher:
        #                                 CREATE TEXT INDEX word_lookup FOR (n:Word) ON (n.name)

        assert NeoSchema.is_valid_class_name(content_item_class_name), \
            "initialize_schema(): a non-empty string is required for argument `content_item_class_name`"

        if NeoSchema.class_name_exists(content_item_class_name):
            content_item_class_id = NeoSchema.get_class_internal_id(class_name=content_item_class_name)
        else:
            content_item_class_id, _ = NeoSchema.create_class(name=content_item_class_name, strict=False)


        indexer_class_id, _ = NeoSchema.create_class(name="Indexer", strict=True)

        NeoSchema.create_class_with_properties(name="Word", strict=True,
                                               properties=["name"],
                                               class_to_link_to="Indexer", link_name="occurs", link_dir="OUT")

        NeoSchema.create_class_relationship(from_class=content_item_class_id, to_class=indexer_class_id, rel_name="has_index")



    @classmethod
    def new_indexing(cls, internal_id :int, unique_words :Set[str], to_lower_case=True) -> None:
        """
        Used to create a new index in the database
        for the (single) specified data node that represents a "Content Item".
        The indexing will link that Content Item to the given list of unique words.

        An Exception is raised if the "Indexer" node already exists

        Details:
        1) Create a data node of type "Indexer",
            with inbound relationships named "occurs" from "Word" data nodes
            (pre-existing or newly-created as needed)
            for all the words in the given list
        2) create a relationship named "has_index" from an existing "Content Item" data node
            to the new "Indexer" node

        :param internal_id:  The internal database ID of an existing data node
                                    that represents a "Content Item" (not necessarily with that Schema name)
        :param unique_words:    A list of strings containing unique words "worthy" of indexing
                                    - for example as returned by extract_unique_good_words()
        :param to_lower_case:   If True, all text is converted to lower case
        :return:                None
        """
        # TODO: consider combining new_indexing() and update_indexing()

        indexer_id = cls.get_indexer_node_id(internal_id)
        assert indexer_id is None, \
            f"new_indexing(): an index ALREADY exists for the given Content Item node (internal id {internal_id})"

        # Create a data node of type "Indexer", and link it up to the passed Content Item data node
        indexer_id = NeoSchema.add_data_node_with_links(class_name ="Indexer",
                                                        links =[{"internal_id": internal_id, "rel_name": "has_index",
                                                                  "rel_dir": "IN"}])

        cls.add_words_to_index(indexer_id=indexer_id, unique_words=unique_words, to_lower_case=to_lower_case)



    @classmethod
    def add_words_to_index(cls, indexer_id :int, unique_words :Set[str], to_lower_case=True) -> int:
        """
        Add to the database "Word" nodes for all the given words, unless already present.
        Then link all the "Word" nodes, both the found and the newly-created ones,
        to the passed "Indexer" node with an "occurs" relationships

        :param indexer_id:      Internal database ID of an existing "Indexer" data node
                                    used to hold all the "occurs" relationships
                                    to the various Word nodes.
                                    If not present, an Exception gets raised.
        :param unique_words:    Set of strings, with unique words for the index
        :param to_lower_case:   If True, all text is converted to lower case
        :return:                The number of new "Word" Data Notes that were created
        """
        # Validate indexer_id
        CypherUtils.assert_valid_internal_id(indexer_id)
        q = '''
            MATCH (i :Indexer)-[:SCHEMA]->(:CLASS {name: "Indexer"})
            WHERE id(i) = $indexer_id 
            RETURN count(i) AS number_of_nodes
            '''
        result = cls.db.query(q, data_binding={"indexer_id": indexer_id},
                              single_cell="number_of_nodes")
        assert result > 0, \
            f"add_words_to_index(): there's no Indexer node with internal ID {indexer_id}"


        if to_lower_case:
            unique_words = list(map(str.lower, unique_words))
        else:
            unique_words = list(unique_words)

        # The following query locates (or creates if not found) a "Word" data node
        # for each word in the list unique_words,
        # and then (as needed) it links it up to the "Indexer" node (ind), with an "occurs" relationship,
        # and to the "Class" node named "Word" (wcl), with a "SCHEMA" relationship.
        # Note: any already-existing "Word" data node ALREADY possess a link to the common Class node;
        #       hence, the line   "MERGE (w :`Word` {name : word})-[:SCHEMA]->(wcl)"
        q = '''
            MATCH (ind :`Indexer`), (wcl :`CLASS` {name: "Word"})
            WHERE id(ind) = $indexer_id
            WITH ind, wcl
            UNWIND $word_list AS word
            MERGE (w :`Word` {name : word})-[:SCHEMA]->(wcl)
            MERGE (ind)<-[:occurs]-(w)
            '''

        data_binding = {"indexer_id": indexer_id, "word_list": unique_words}
        #print("add_words_to_index(): about to run the query to update the index")
        result = cls.db.update_query(q, data_binding)
        #print(result)
        # EXAMPLE of result:
        #   {'labels_added': 3, '_contains_updates': True, 'relationships_created': 6, 'nodes_created': 3, 'properties_set': 3, 'returned_data': []}
        number_word_nodes_added = result.get('nodes_created', 0)
        number_word_nodes_found = len(unique_words) - number_word_nodes_added

        assert result.get('labels_added', 0) == number_word_nodes_added, \
            "add_words_to_index(): internal consistency error; " \
            "the number of labels created should have equaled the number of nodes created"

        assert result.get('properties_set', 0) == number_word_nodes_added, \
            "add_words_to_index(): internal consistency error; " \
            "the number of properties being set should have equaled the number of nodes created"

        # To determine a lower and upper bound on the the number of relationships added,
        # consider that ech newly-create Word data node adds 2 relationships (to the "Word" Class and to the "Indexer" node);
        # by contrast, each found node only adds at most 1 relationship (to the "Indexer" node)
        lb = 2 * number_word_nodes_added
        ub =  lb + + number_word_nodes_found
        assert lb <= result.get('relationships_created', 0) <= ub, \
            f"add_words_to_index(): internal consistency error; " \
            f"the number of relationships_created should have been between {lb} and {ub}, inclusive; " \
            f"instead of result.get('relationships_created', 0)"

        return number_word_nodes_added



    @classmethod
    def update_indexing(cls, content_uri :int, unique_words :Set[str], to_lower_case=True) -> None:
        """
        Used to update an index, linking the given list of unique words
        to the specified "Indexer" data node, which was created by a call to new_indexing()
        at the time the index was first created.
        
        From the given data node of type "Indexer",
        add inbound relationships named "occurs" from "Word" data nodes (pre-existing or newly-created)
        for all the words in the given list.
        Also, create a relationship named "has_index" from an existing "Content Item" data node to the new "Indexer" node.

        Note: if no index exist, an Exception is raised

        :param content_uri: The internal database ID of an existing "Content Item" data node
        :param unique_words:    A list of strings containing unique words
                                    - for example as returned by extract_unique_good_words()
        :param to_lower_case:   If True, all text is converted to lower case
        :return:                None
        """
        indexer_id = cls.get_indexer_node_id(content_uri)
        assert indexer_id is not None, \
                    f"update_indexing(): unable to find an index for the given Content Item " \
                    f" (internal id {content_uri}).  Did you first create an index for it?"

        # Sever all the existing "occurs" relationships to the "Indexer" data node
        # i.e. give a "clean slate" to the "Indexer" data node
        NeoSchema.remove_multiple_data_relationships(node_id=indexer_id, rel_name="occurs", rel_dir="IN", labels="Word")

        try:
            #print("update_indexing(): about to calling add_words_to_index()")
            cls.add_words_to_index(indexer_id=indexer_id, unique_words=unique_words, to_lower_case=to_lower_case)
            #print("update_indexing(): returned from call to add_words_to_index()")
        except Exception as ex:
            err_details = f"Failure in FullTextIndexing.add_words_to_index().  {exceptions.exception_helper(ex)}"
            raise Exception(err_details)



    @classmethod
    def get_indexer_node_id(cls, internal_id :int) -> Union[int, None]:
        """
        Retrieve and return the internal database ID of the "Indexer" data node
        associated to the given Content Item data node.
        If not found, None is returned

        :param internal_id: The internal database ID of an existing data node
                                    (either of Class "Content Item", or of a Class that is ultimately
                                    an INSTANCE_OF a "Content Item" Class)
        :return:                The internal database ID of the corresponding "Indexer" data node.
                                    If not found, None is returned
        """
        # Prepare a Cypher query
        q = '''
            MATCH (ci)-[:has_index]->(i:Indexer)-[:SCHEMA]->(:CLASS {name: "Indexer"})
            WHERE id(ci) = $content_uri
            RETURN id(i) AS indexer_id
            '''

        return cls.db.query(q, data_binding={"content_uri": internal_id},
                            single_cell="indexer_id")       # Note: will be None if no record found



    @classmethod
    def remove_indexing(cls, content_uri :int) -> None:
        """
        Drop the "Indexer" node linked to the given Content Item node.
        If no index exists, an Exception is raised

        :param content_uri: The internal database ID of an existing "Content Item" data node
        :return:                None
        """
        indexer_id = cls.get_indexer_node_id(content_uri)

        assert indexer_id is not None, \
            f"remove_indexing(): unable to find an index for the given Content Item node" \
            f" (internal id {content_uri}).  Maybe you already removed it?"

        NeoSchema.delete_data_node(node_id=indexer_id, labels="Indexer", class_node="Indexer")



    @classmethod
    def number_of_indexed_words(cls, internal_id=None, uri=None) -> int:
        """
        Determine and return the number of words attached to the index
        of the given data node (typically of a Class representing "Content Item" ,
        or instance thereof, such as "Document" or "Note")

        :param internal_id: The internal database ID of an existing Content Item data node
        :param uri:         Alternate way to specify the Content Item data node, with a string URI
        :return:            The number of indexed words associated to the above node
        """
        if internal_id is not None:
            assert type(internal_id) == int, "number_of_indexed_words(): argument `internal_id` must be an integer"
            clause = "WHERE id(ci) = $internal_id"
            data_binding = {"internal_id": internal_id}
        elif uri:
            assert type(uri) == str, "number_of_indexed_words(): argument `uri` must be a string"
            clause = "WHERE ci.uri = $uri"
            data_binding = {"uri": uri}
        else:
            raise Exception("number_of_indexed_words(): at least one argument must be specified")

        q = f'''
            MATCH (wc :CLASS {{name:"Word"}})<-[:SCHEMA]-
            (w :Word)-[:occurs]->(i :Indexer)<-[:has_index]-(ci) 
            {clause}
            RETURN count(w) AS word_count
            '''
        return cls.db.query(q, data_binding=data_binding, single_cell="word_count")



    @classmethod
    def most_common_words_NOT_YET_USED(cls) -> [str]:    # TODO: finish implementing
        """
        Return the 200 or so most common words in the current index
        """
        q = '''
            MATCH (w:Word)-[:occurs]->(i:Indexer) 
            RETURN w.name, count(i) AS c
            ORDER by c DESC LIMIT 200
        '''

        return ["Not yet implemented"]



    @classmethod
    def word_occurrence_NOT_YET_USED(cls, word: str) -> int:    # TODO: finish implementing
        """
        Return the usage count of the given word in the current index
        :param word:
        :return:
        """
        q = '''
        MATCH (w:Word {name:$word})-[:occurs]->(i:Indexer) 
        RETURN count(i) AS word_occurrence
        '''

        return 0



    @classmethod
    def index_size_NOT_YET_USED(cls):    # TODO: finish implementing
        """
        Return the number of content items current being indexed

        :return:
        """
        q = '''
            MATCH (n:Indexer) RETURN count(n)
            '''



    @classmethod
    def number_word_NOT_YET_USED(cls):    # TODO: finish implementing
        """
        Return the number of words current being used in the indexing

        :return:
        """
        q = '''
            MATCH (n:Word) RETURN count(n)
            '''


    @classmethod
    def unused_word_NOT_YET_USED(cls):    # TODO: finish implementing
        """
        Return a list of words in the indexing that lack any usage

        :return:
        """
        pass



    @classmethod
    def rebuild_index_NOT_YET_USED(cls, directory :str):    # TODO: finish implementing
        """
        Rebuild the index from all the (text or HTML) files in the given directory.

        TODO: a version that looks up that list from the database

        :param directory:     EXAMPLE:  "D:/tmp/transfer"  (Use forward slashes even on Windows!)
        :return:
        """
        import os
        from brainannex.media_manager import MediaManager

        file_list = os.listdir(directory)
        #print(f"Total number of files: {len(file_list)}")

        # Index the content of all the files
        i = 1
        for filename in file_list:
            #print(f"\n {i} -------------------------\n", filename)
            (basename, suffix) = os.path.splitext(filename)
            q = f"MATCH (n:Notes) WHERE n.basename='{basename}' AND n.suffix='htm' RETURN ID(n) AS node_int_id"
            node_int_id = cls.db.query(q, single_cell="node_int_id")
            #print("    node's integer ID: ", node_int_id)
            path = "TBA"
            file_contents = MediaManager.get_from_text_file(filename, path)
            #print(file_contents)
            word_list = FullTextIndexing.extract_unique_good_words(file_contents)
            #print(word_list)
            FullTextIndexing.new_indexing(internal_id= node_int_id, unique_words = word_list)
            i += 1




    #########################   SEARCHING   #########################

    @classmethod
    def search_word(cls, word :str, all_properties=False,
                    restrict_search=None, search_category=None) -> Union[List[int], List[dict]]:
        """
        Look up any stored words that contains the requested string
        (ignoring case and leading/trailing blanks.)

        Then locate the Content nodes that are indexed by any of those words.

        Return a (possibly empty) list of either the internal database ID's of all the found nodes,
        or a list of their full properties.

        :param word:            A string, typically containing a word or word fragment;
                                    case is ignored, and so are leading/trailing blanks
        :param all_properties:  If True, the properties of the located nodes are returned
                                    alongside their internal database ID's.
                                    Default is False: only return the internal database ID's
        :param restrict_search: If None or an empty list, ignored;
                                    otherwise, it should be a list of internal database ID's to which
                                    the search is to be limited to
        :param search_category: (OPTIONAL) URI of Category.  If supplied, all searching will
                                    be limited to Content Items in this Category
                                    or in any of its sub-categories

        :return:        If all_properties is False,
                            a (possibly empty) list of the internal database ID's
                            of all the found nodes
                        If all_properties is True,
                            a (possibly empty) list of dictionaries with all the data
                            of all the found nodes; each dict contain all of the nodes' attributes,
                            plus keys called 'internal_id' and 'neo4j_labels'
                            EXAMPLE: [{'filename': 'My_Document.pdf', 'internal_id': 66, 'neo4j_labels': ['Content Item']}]
        """
        clean_term = word.strip()   # Zap leading/trailing blanks

        if clean_term == "":
            # This is done as special handling because a blank string would match any word!
            return []

        if all_properties:
            return_statement = "RETURN DISTINCT ci"
        else:
            return_statement = "RETURN DISTINCT id(ci) AS content_id"


        where_additional_clause = ""
        data_binding = {}
        additional_matching = ""

        if restrict_search:
            #print("Restricting search to Content Items with internal ID's: ", restrict_search)
            where_additional_clause += " AND id(ci) IN $restrict_search"
            data_binding["restrict_search"] = restrict_search

        if search_category:
            #print("Restricting search to Content Items under Category with URI: ", search_category)
            additional_matching = "-[:BA_in_category]->(:Categories)-[:BA_subcategory_of*0..]->(cat:Categories)"
            where_additional_clause += " AND cat.uri = $search_category"
            data_binding["search_category"] = search_category


        q = f'''
            MATCH (:CLASS {{name:"Word"}})<-[:SCHEMA]-
            (w:Word)-[:occurs]->(:Indexer)<-[:has_index]-(ci)
            {additional_matching}
            WHERE w.name CONTAINS toLower('{clean_term}')
            {where_additional_clause}
            {return_statement} 
            '''

        #cls.db.debug_query_print(q=q, data_binding=data_binding, method="search_word")


        if all_properties:
            result = cls.db.query_extended(q, data_binding=data_binding, flatten=True)
        else:
            result = cls.db.query(q, data_binding=data_binding, single_column="content_id")

        return result



from brainannex.PLUGINS.documents import Documents
from brainannex.PLUGINS.images import Images
from brainannex.PLUGINS.notes import Notes


def default_folder(class_name :str) -> str:
    """
    Fetch the name for the default folder used for media content associated to the given class

    :param class_name:
    :return:            A folder name, with no slashes.  EXAMPLE: "documents"
    """
    if class_name == "Document":
        return Documents.default_folder()

    if class_name == "Images":  # TODO: change to singular
        return Images.default_folder()

    if class_name == "Notes":   # TODO: change to singular
        return Notes.default_folder()

    raise Exception(f"plugin_support.default_folder(): unknown or missing value for "
                    f"`class_name` ({class_name})")



def all_default_folders() -> dict:
    """
    Return a dict mapping a Class name to its designated default folder

    :return:    EXAMPLE: {"Document": "documents", "Image": "images", "Note": "notes"}
    """
    d = {}

    d["Document"] = Documents.default_folder()
    d["Images"] = Images.default_folder()           # TODO: change to singular
    d["Notes"] = Notes.default_folder()             # TODO: change to singular

    return d



def is_media_class(class_name :str) -> bool:
    """
    Return True if the given Class is an INSTANCE_OF the "Media" Class

    :param class_name:  Name of a Schema class
    :return:            True if the given Class is an INSTANCE_OF the "Media" Class
    """
    # TODO: for now hardwired; ought to instead query the Schema, to discover if the given Class
    #       is an INSTANCE_OF the "Media" class
    return class_name in ["Document", "Images", "Notes"]    # TODO: change to singular
